"""IdfClass utilities."""

import hashlib
import io
import os
from collections import OrderedDict
from io import StringIO
from typing import List, Union

from packaging.version import Version


def hash_model(idfname, **kwargs):
    """Hash a file or IDF model.

    Return it as a string. Will also hash the :func:`eppy.runner.run_functions.run()`
    arguments so that correct results are returned when different run arguments are
    used.

    Todo:
        Hashing should include the external files used an idf file. For example,
        if a model uses a csv file as an input and that file changes, the
        hashing will currently not pickup that change. This could result in
        loading old results without the user knowing.

    Args:
        idfname (str or IDF): path of the idf file or the IDF model itself.
        kwargs: kwargs to serialize in addition to the file content.

    Returns:
        str: The digest value as a string of hexadecimal digits
    """
    from .idf import IDF

    if kwargs:
        # Before we hash the kwargs, remove the ones that don't have an impact on
        # simulation results and so should not change the cache dirname.
        no_impact = ["keep_data", "keep_data_err", "return_idf", "return_files"]
        for argument in no_impact:
            _ = kwargs.pop(argument, None)

        # sorting keys for serialization of dictionary
        kwargs = OrderedDict(sorted(kwargs.items()))

    # create hasher
    hasher = hashlib.md5()
    if isinstance(idfname, StringIO):
        idfname.seek(0)
        buf = idfname.read().encode("utf-8")
    elif isinstance(idfname, IDF):
        buf = idfname.idfstr().encode("utf-8")
        if idfname.name:
            hasher.update(idfname.name.encode("utf-8"))
            # hash idfname.basename in case file content is identical with another
            # file with a different name
    else:
        with open(idfname, "rb") as afile:
            buf = afile.read()
    hasher.update(buf)

    # Hashing the kwargs as well
    for k, v in kwargs.items():
        if isinstance(v, (str, bool)):
            hasher.update(v.__str__().encode("utf-8"))
        elif isinstance(v, list):
            # include files are Paths
            for item in v:
                with open(item, "rb") as f:
                    buf = f.read()
                    hasher.update(buf)
    return hasher.hexdigest()


def get_idf_version(file: Union[str, io.StringIO], doted=True, encoding=None):
    """Get idf version quickly by reading first few lines of idf file containing
    the 'VERSION' identifier

    Args:
        file (str or StringIO): Absolute or relative Path to the idf file
        doted (bool, optional): Wheter or not to return the version number
        encoding (str): The encoding used to open the file.

    Returns:
        str: the version id
    """
    import re

    if isinstance(file, io.StringIO):
        file.seek(0)
        txt = file.read()
    else:
        with open(file, encoding=encoding) as f:
            txt = f.read()

    versions: List = re.findall(r"(?s)(?<=Version,).*?(?=;)", txt, re.IGNORECASE)
    for v in versions:
        version = Version(v.strip())
        if doted:
            return f"{version.major}.{version.minor}.{version.micro}"
        return f"{version.major}-{version.minor}-{version.micro}"


def getoldiddfile(versionid):
    """find the IDD file of the E+ installation E+ version 7 and earlier have
    the idd in /EnergyPlus-7-2-0/bin/Energy+.idd

    Args:
        versionid:
    """
    import eppy

    vlist = versionid.split(".")
    if len(vlist) == 1:
        vlist = vlist + ["0", "0"]
    elif len(vlist) == 2:
        vlist = vlist + ["0"]
    ver_str = "-".join(vlist)
    eplus_exe, _ = eppy.runner.run_functions.install_paths(ver_str)
    eplusfolder = os.path.dirname(eplus_exe)
    iddfile = "{}/bin/Energy+.idd".format(eplusfolder)
    return iddfile



"""Module for parsing EnergyPlus SQLite result files into DataFrames."""
import logging
from datetime import timedelta
from sqlite3 import connect
from typing import List, Optional, Sequence, Union

import numpy as np
import pandas as pd
from energy_pandas import EnergyDataFrame
from pandas import to_datetime
from path import Path
from typing_extensions import Literal

from archetypal.utils import log

_REPORTING_FREQUENCIES = Literal[
    "HVAC System Timestep",
    "Zone Timestep",
    "Hourly",
    "Daily",
    "Monthly",
    "Run Period",
]


class SqlOutput:
    """Represents a single output from the Sql file."""

    __slots__ = (
        "_file_path",
        "output_name",
        "reporting_frequency",
    )

    def __init__(self, file_path, output_name, reporting_frequency):
        self._file_path = file_path
        self.output_name = output_name
        self.reporting_frequency = reporting_frequency

    def values(self, environment_type: int = 3, units: str = None) -> EnergyDataFrame:
        """Get the time series values as an EnergyDataFrame.

        Args:
            environment_type (int): The environment type (1 = Design Day, 2 = Design
                Run Period, 3 = Weather Run Period) for the series.
            units (str): Convert original values to another unit. The original unit
                is detected automatically and a dimensionality check is performed.

        Returns:
            (EnergyDataFrame): The time series as an EnergyDataFrame.
        """
        cols = (
            "ReportDataDictionaryIndex, IndexGroup, KeyValue, Name, "
            "Units, ReportingFrequency"
        )
        query = f"""
            SELECT {cols} 
            FROM ReportDataDictionary 
            WHERE Name=@output_name 
            AND ReportingFrequency=@reporting_frequency;
        """
        with connect(self._file_path) as conn:
            header_rows = pd.read_sql(
                query,
                conn,
                params={
                    "output_name": self.output_name,
                    "reporting_frequency": self.reporting_frequency,
                },
            )
            header_rows.set_index("ReportDataDictionaryIndex", inplace=True)
            # extract all data of the relevant type from ReportData
            rel_indices = tuple(header_rows.index.to_list())
            data = _extract_timeseries(conn, environment_type, header_rows, rel_indices)

            if units is not None:
                data = data.to_units(units)

            return data


class _SqlOutputs:
    """Represents all the available outputs from the Sql file."""

    def __init__(self, file_path: str, available_outputs: List[tuple]):
        self._available_outputs = available_outputs
        self._properties = {}

        for output, reporting_frequency in self._available_outputs:
            name = (
                output.replace(":", "__").replace(" ", "_") + f"_{reporting_frequency}"
            )
            self._properties[name] = SqlOutput(file_path, output, reporting_frequency)
            setattr(self, name, self._properties[name])

    def __getitem__(self, meter_name):
        """Get item by key."""
        return self._properties[meter_name]


class Sql:
    """Object for parsing EnergyPlus SQLite result files into DataFrames.

    Args:
        file_path: Full path to an SQLite file that was generated by EnergyPlus.

    Properties:
        * file_path
        * available_outputs
        * zone_info
        * environment_periods
    """

    _reporting_frequencies = (
        "HVAC System Timestep",
        "Zone Timestep",
        "Hourly",
        "Daily",
        "Monthly",
        "Run Period",
    )

    def __init__(self, file_path):
        """Initialize SQLiteResult"""
        assert Path(file_path).exists(), "No file was found at {}".format(file_path)
        self._file_path = file_path

        # values to be computed as soon as they are requested
        self._available_outputs = None
        self._zone_info = None
        self._environment_periods = None
        self._tabular_data_keys = None
        self._outputs = None
        self._surfaces_table = None
        self._constructions_table = None

    @property
    def file_path(self):
        """Get the path to the .sql file."""
        return self._file_path

    @property
    def tabular_data_keys(self):
        """Get tuples of (ReportName, TableName, ReportForString) from tabular data."""
        if self._tabular_data_keys is None:
            with connect(self.file_path) as conn:
                query = "SELECT DISTINCT ReportName, TableName, ReportForString FROM TabularDataWithStrings"
                data = conn.execute(query)
                self._tabular_data_keys = data.fetchall()
        return self._tabular_data_keys

    @property
    def available_outputs(self) -> List[tuple]:
        """Get tuples (OutputName, ReportingFrequency) that can be requested.

        Any of these outputs when input to data_collections_by_output_name will
        yield a result with data collections.
        """
        if self._available_outputs is None:
            self._available_outputs = self._extract_available_outputs()
        return self._available_outputs

    @property
    def outputs(self):
        if self._outputs is None:
            self._outputs = _SqlOutputs(self.file_path, self.available_outputs)
        return self._outputs

    @property
    def zone_info(self):
        """Get a list of strings for available timeseries outputs that can be requested.

        Any of these outputs when input to data_collections_by_output_name will
        yield a result with data collections.
        """
        if self._zone_info is None:
            self._zone_info = self._extract_zone_info()
        return self._zone_info

    @property
    def surfaces_table(self):
        if self._surfaces_table is None:
            self._surfaces_table = self._extract_surfaces_table()
        return self._surfaces_table

    @property
    def constructions_table(self):
        if self._constructions_table is None:
            self._constructions_table = self._extract_constructions_table()
        return self._constructions_table

    @property
    def environment_periods(self):
        """Get a list of environment periods for the simulation run periods.

        EnvironmentType: An enumeration of the environment type. (1 = Design Day,
            2 = Design Run Period, 3 = Weather Run Period).
        """
        if self._environment_periods is None:
            self._environment_periods = self._extract_environment_periods()
        return self._environment_periods

    def full_html_report(self):
        """Get the html report as a dictionary of DataFrames.

        The dictionary keys are tuples of ("ReportName", "TableName",
        "ReportForString").
        """
        with connect(self.file_path) as conn:
            cols = (
                "ReportName, TableName, ReportForString, ColumnName, RowName, "
                "Units, Value"
            )
            query = f"SELECT {cols} FROM TabularDataWithStrings"
            data = pd.read_sql(query, conn)

        data.RowName = data.RowName.replace({"": np.NaN, "-": np.NaN})
        data.dropna(subset=["RowName"], inplace=True)

        all_df = {}
        for name, df in data.groupby(["ReportName", "TableName", "ReportForString"]):
            try:
                pivoted = df.pivot(
                    columns=["ColumnName", "Units"], index="RowName", values="Value"
                )
            except ValueError:
                # Cannot pivot; return long form
                pivoted = df
            # apply to_numeric column-wise
            pivoted = pivoted.apply(pd.to_numeric, errors="ignore")

            # insert in final dict.
            all_df[name] = pivoted

        return all_df

    def timeseries_by_name(
        self,
        variable_or_meter: Union[str, Sequence],
        reporting_frequency: Union[_REPORTING_FREQUENCIES] = "Hourly",
        environment_type: Union[Literal[1, 2, 3]] = 3,
    ) -> EnergyDataFrame:
        """Get an EnergyDataFrame for specified meters and/or variables.

        The returned DataFrame has a column MultiIndex with levels ["IndexGroup",
        "KeyValue", "Name"]. KeyValue corresponds to the zone name for variables
        while KeyValue is None for meters.

        Note that if an output name is not not available for the reporting
        frequency or the environment type, the DataFrame can be empty.

        Args:
            variable_or_meter (str or list): The name of an EnergyPlus output meter or
                variable to be retrieved from the SQLite result file. This can also be an
                array of output names for which all data collections should be retrieved.
            reporting_frequency (str): The reporting interval. One of ("HVAC System
                Timestep", "Zone Timestep", "Hourly", "Daily", "Monthly" or "Run
                Period"
            environment_type (int): The environment type. (1 = Design Day, 2 = Design
                Run Period, 3 = Weather Run Period). Default = 3.

        Returns:
            EnergyDataFrame: An EnergyDataFrame with the variable_or_meter as columns.
        """
        reporting_frequency = reporting_frequency.title()
        assert (
            reporting_frequency in Sql._reporting_frequencies
        ), f"reporting_frequency is not one of {Sql._reporting_frequencies}"
        with connect(self.file_path) as conn:
            cols = "ReportDataDictionaryIndex, IndexGroup, KeyValue, Name, Units, ReportingFrequency"
            if isinstance(variable_or_meter, str):  # assume it's a single output
                if (
                    variable_or_meter,
                    reporting_frequency,
                ) not in self.available_outputs:
                    log(
                        f"{(variable_or_meter, reporting_frequency)} not "
                        f"an available output in the Sql file.",
                        level=logging.WARNING,
                    )
                query = f"""
                        SELECT {cols} 
                        FROM ReportDataDictionary 
                        WHERE Name=@output_name 
                        AND ReportingFrequency=@reporting_frequency;
                        """
                header_rows = pd.read_sql(
                    query,
                    conn,
                    params={
                        "output_name": variable_or_meter,
                        "reporting_frequency": reporting_frequency,
                    },
                )
            elif len(variable_or_meter) == 1:  # assume it's a list
                query = f"""
                        SELECT {cols} 
                        FROM ReportDataDictionary 
                        WHERE Name=@output_name 
                        AND ReportingFrequency=@reporting_frequency;
                        """
                header_rows = pd.read_sql(
                    query,
                    conn,
                    params={
                        "output_name": variable_or_meter[0],
                        "reporting_frequency": reporting_frequency,
                    },
                )
            else:  # assume it is a list of outputs
                query = f"""
                        SELECT {cols} 
                        FROM ReportDataDictionary 
                        WHERE Name IN {tuple(variable_or_meter)}
                        AND ReportingFrequency=@reporting_frequency;"""
                header_rows = pd.read_sql(
                    query,
                    conn,
                    params={
                        "reporting_frequency": reporting_frequency,
                    },
                )
            # if nothing was found, return an empty DataFrame
            if len(header_rows) == 0:
                return EnergyDataFrame([])
            else:
                header_rows.set_index("ReportDataDictionaryIndex", inplace=True)

            # extract all data of the relevant type from ReportData
            rel_indices = tuple(header_rows.index.to_list())
            data = _extract_timeseries(conn, environment_type, header_rows, rel_indices)
        log(f"collected data for {variable_or_meter}")
        return data

    def tabular_data_by_name(
        self, report_name: str, table_name: str, report_for_string: Optional[str] = None
    ) -> pd.DataFrame:
        """Get (ReportName, TableName) data as DataFrame.

        Args:
            report_name (str): The name of the report.
            table_name (str): The name of the table in the report.
            report_for_string (str): The “For” string.

        Returns:
            (pd.DataFrame): A DataFrame.
        """
        with connect(self.file_path) as conn:
            cols = "RowName, ColumnName, Value, Units"
            query = f"""
                SELECT {cols} FROM TabularDataWithStrings 
                WHERE 
                    (@report_name IS NULL OR ReportName=@report_name)
                AND 
                    (@table_name IS NULL OR TableName=@table_name)
                AND 
                    (@report_for_string IS NULL OR ReportForString=@report_for_string);
            """
            data = pd.read_sql(
                query,
                conn,
                params={
                    "report_name": report_name,
                    "table_name": table_name,
                    "report_for_string": report_for_string,
                },
            )
            try:
                pivoted = data.pivot(
                    index="RowName", columns=["ColumnName", "Units"], values="Value"
                )
            except ValueError:
                # Cannot pivot; return long-form DataFrame
                pivoted = data
                log(
                    f"{(report_name, table_name, report_name)} cannot be "
                    f"pivoted as RowName and ColumnName. The long-form "
                    f"DataFrame has been returned.",
                    level=logging.WARNING,
                )
            pivoted = pivoted.apply(pd.to_numeric, errors="ignore")
        return pivoted

    def _extract_available_outputs(self) -> List:
        """Extract the list of all available outputs from the SQLite file."""
        with connect(self.file_path) as conn:
            cols = "Name, ReportingFrequency"
            query = f"SELECT DISTINCT {cols} FROM ReportDataDictionary"
            data = conn.execute(query)
            return data.fetchall()

    def _extract_zone_info(self):
        """Extract the Zones table from the SQLite file."""
        with connect(self.file_path) as conn:
            query = "SELECT * from Zones"
            df = pd.read_sql(query, conn).set_index("ZoneIndex")
        return df

    def _extract_surfaces_table(self):
        """Extract the Zones table from the SQLite file."""
        with connect(self.file_path) as conn:
            query = "SELECT * from Surfaces"
            df = pd.read_sql(query, conn).set_index(["ZoneIndex", "SurfaceIndex"])
        return df

    def _extract_constructions_table(self):
        with connect(self.file_path) as conn:
            query = "SELECT * from Constructions"
            df = pd.read_sql(query, conn).set_index("ConstructionIndex")
        return df

    def _extract_environment_periods(self):
        """Extract the EnvironmentPeriods table from the SQLite file."""
        with connect(self.file_path) as conn:
            query = "SELECT * from EnvironmentPeriods"
            df = pd.read_sql(query, conn).set_index("EnvironmentPeriodIndex")
        return df


def _extract_timeseries(
    conn, environment_type, header_rows, rel_indices
) -> EnergyDataFrame:
    """Extract time series given indices."""
    if len(rel_indices) == 1:
        data = pd.read_sql(
            """SELECT rd.Value,
                      rd.ReportDataDictionaryIndex, 
                      t.Month,
                      t.Day,
                      t.Hour,
                      t.Minute,
                      t.Interval
            FROM ReportData as rd 
                    LEFT JOIN Time As t ON rd.TimeIndex = t.TimeIndex
                    LEFT JOIN EnvironmentPeriods as p ON t.EnvironmentPeriodIndex = p.EnvironmentPeriodIndex
            WHERE ReportDataDictionaryIndex=@rel_indices
            AND (IFNULL(t.WarmupFlag, 0) = @warmup_flag)
            AND p.EnvironmentType = @environment_type
            ORDER BY t.TimeIndex;""",
            conn,
            params={
                "rel_indices": rel_indices[0],
                "warmup_flag": 0,
                "environment_type": environment_type,
            },
        )
    else:
        data = pd.read_sql(
            f"""SELECT rd.Value,
                          rd.ReportDataDictionaryIndex,
                          t.Month,
                          t.Day,
                          t.Hour,
                          t.Minute,
                          t.Interval
                FROM ReportData as rd
                        LEFT JOIN Time As t ON rd.TimeIndex = t.TimeIndex
                        LEFT JOIN EnvironmentPeriods as p ON t.EnvironmentPeriodIndex = p.EnvironmentPeriodIndex
                WHERE ReportDataDictionaryIndex IN {tuple(rel_indices)}
                AND (IFNULL(t.WarmupFlag, 0) = @warmup_flag)
                AND p.EnvironmentType = @environment_type
                ORDER BY rd.ReportDataDictionaryIndex, t.TimeIndex;""",
            conn,
            params={"warmup_flag": 0, "environment_type": environment_type},
        )
    # Join the header_rows on ReportDataDictionaryIndex
    data = data.join(
        header_rows[["IndexGroup", "KeyValue", "Name"]],
        on="ReportDataDictionaryIndex",
    )
    # Pivot the data so that ["Name", "KeyValue"] becomes the column MultiIndex.
    data = data.pivot(
        index=["Month", "Day", "Hour", "Minute", "Interval"],
        columns=["IndexGroup", "KeyValue", "Name"],
        values="Value",
    )
    # reset the index to prepare the DatetimeIndex
    date_time_names = data.index.names
    data.reset_index(inplace=True)
    index = to_datetime(
        {
            "year": 2018,
            "month": data.Month,
            "day": data.Day,
            "hour": data.Hour,
            "minute": data.Minute,
        }
    )
    # Adjust timeindex by timedelta
    index -= data["Interval"].apply(lambda x: timedelta(minutes=x))
    index = pd.DatetimeIndex(index, freq="infer")
    # get data
    data = data.drop(columns=date_time_names, level="IndexGroup")
    data.index = index
    # Create the EnergyDataFrame and set the units using dict
    data = EnergyDataFrame(data)
    data.units = header_rows.set_index(["IndexGroup", "KeyValue", "Name"])[
        "Units"
    ].to_dict()
    return data



################################################################################
# Module: __init__.py
# Description: Archetypal: Retrieve, construct and analyse building archetypes
# License: MIT, see full license in LICENSE.txt
# Web: https://github.com/samuelduchesne/archetypal
################################################################################
import logging as lg
from pathlib import Path
from typing import Literal, List, Optional, Any

from energy_pandas.units import unit_registry

# Version of the package
from pkg_resources import get_distribution, DistributionNotFound
from pydantic import (
    BaseSettings,
    Field,
    validator,
    DirectoryPath,
)

# don't display futurewarnings

import warnings

warnings.simplefilter(action="ignore", category=FutureWarning)
warnings.simplefilter(action="ignore", category=UserWarning)


class ZoneWeight(object):
    """Zone weights for Umi Templates"""

    weight_attr = {0: "area", 1: "volume"}

    def __init__(self, n=0):
        self._weight_attr = self.weight_attr[n]

    def __str__(self):
        return self.get_weight_attr()

    def get_weight_attr(self):
        return self._weight_attr

    def set_weigth_attr(self, weight):
        if weight not in self.weight_attr.values():
            i = len(self.weight_attr) + 1
            self.weight_attr[i] = weight
        self._weight_attr = weight


class Settings(BaseSettings):
    class Config:
        arbitrary_types_allowed = True
        validate_assignment = True

    data_folder: Path = Field("data", env="ARCHETYPAL_DATA")
    logs_folder: Path = Field("logs", env="ARCHETYPAL_LOGS")
    imgs_folder: Path = Field("images", env="ARCHETYPAL_IMAGES")
    cache_folder: Path = Field("cache", env="ARCHETYPAL_CACHE")

    # cache server responses
    cache_responses: bool = Field(False, env="ARCHETYPAL_CACHE_RESPONSES")

    # Debug behavior
    debug = Field(False, env="ARCHETYPAL_DEBUG")

    # write log to file and/or to console
    log_file: bool = Field(False)
    log_console: bool = Field(False)
    log_notebook: bool = Field(False)
    log_level: Literal[0, 10, 20, 30, 40, 50] = Field(
        lg.INFO, env="ARCHETYPAL_LOG_LEVEL"
    )
    log_name: str = Field("archetypal", env="ARCHETYPAL_LOG_NAME")
    log_filename: str = Field("archetypal")

    # usual idfobjects
    useful_idf_objects: List[str] = [
        "WINDOWMATERIAL:GAS",
        "WINDOWMATERIAL:GLAZING",
        "WINDOWMATERIAL:SIMPLEGLAZINGSYSTEM",
        "MATERIAL",
        "MATERIAL:NOMASS",
        "CONSTRUCTION",
        "BUILDINGSURFACE:DETAILED",
        "FENESTRATIONSURFACE:DETAILED",
        "SCHEDULE:DAY:INTERVAL",
        "SCHEDULE:WEEK:DAILY",
        "SCHEDULE:YEAR",
    ]

    # List of Available SQLite Tables
    # Ref: https://bigladdersoftware.com/epx/docs/8-3/output-details-and-examples
    # /eplusout.sql.html#schedules-table

    available_sqlite_tables: dict = dict(
        ComponentSizes={"PrimaryKey": ["ComponentSizesIndex"], "ParseDates": []},
        ConstructionLayers={"PrimaryKey": ["ConstructionIndex"], "ParseDates": []},
        Constructions={"PrimaryKey": ["ConstructionIndex"], "ParseDates": []},
        Materials={"PrimaryKey": ["MaterialIndex"], "ParseDates": []},
        NominalBaseboardHeaters={
            "PrimaryKey": ["NominalBaseboardHeaterIndex"],
            "ParseDates": [],
        },
        NominalElectricEquipment={
            "PrimaryKey": ["NominalElectricEquipmentIndex"],
            "ParseDates": [],
        },
        NominalGasEquipment={
            "PrimaryKey": ["NominalGasEquipmentIndex"],
            "ParseDates": [],
        },
        NominalHotWaterEquipment={
            "PrimaryKey": ["NominalHotWaterEquipmentIndex"],
            "ParseDates": [],
        },
        NominalInfiltration={
            "PrimaryKey": ["NominalInfiltrationIndex"],
            "ParseDates": [],
        },
        NominalLighting={"PrimaryKey": ["NominalLightingIndex"], "ParseDates": []},
        NominalOtherEquipment={
            "PrimaryKey": ["NominalOtherEquipmentIndex"],
            "ParseDates": [],
        },
        NominalPeople={"PrimaryKey": ["NominalPeopleIndex"], "ParseDates": []},
        NominalSteamEquipment={
            "PrimaryKey": ["NominalSteamEquipmentIndex"],
            "ParseDates": [],
        },
        NominalVentilation={
            "PrimaryKey": ["NominalVentilationIndex"],
            "ParseDates": [],
        },
        ReportData={"PrimaryKey": ["ReportDataIndex"], "ParseDates": []},
        ReportDataDictionary={
            "PrimaryKey": ["ReportDataDictionaryIndex"],
            "ParseDates": [],
        },
        ReportExtendedData={
            "PrimaryKey": ["ReportExtendedDataIndex"],
            "ParseDates": [],
        },
        RoomAirModels={"PrimaryKey": ["ZoneIndex"], "ParseDates": []},
        Schedules={"PrimaryKey": ["ScheduleIndex"], "ParseDates": []},
        Surfaces={"PrimaryKey": ["SurfaceIndex"], "ParseDates": []},
        SystemSizes={
            "PrimaryKey": ["SystemSizesIndex"],
            "ParseDates": {"PeakHrMin": "%m/%d %H:%M:%S"},
        },
        Time={"PrimaryKey": ["TimeIndex"], "ParseDates": []},
        ZoneGroups={"PrimaryKey": ["ZoneGroupIndex"], "ParseDates": []},
        Zones={"PrimaryKey": ["ZoneIndex"], "ParseDates": []},
        ZoneLists={"PrimaryKey": ["ZoneListIndex"], "ParseDates": []},
        ZoneSizes={"PrimaryKey": ["ZoneSizesIndex"], "ParseDates": []},
        ZoneInfoZoneLists={"PrimaryKey": ["ZoneListIndex"], "ParseDates": []},
        Simulations={
            "PrimaryKey": ["SimulationIndex"],
            "ParseDates": {"TimeStamp": {"format": "YMD=%Y.%m.%d %H:%M"}},
        },
        EnvironmentPeriods={"PrimaryKey": ["EnvironmentPeriodIndex"], "ParseDates": []},
        TabularData={"PrimaryKey": ["TabularDataIndex"], "ParseDates": []},
        Strings={"PrimaryKey": ["StringIndex"], "ParseDates": []},
        StringTypes={"PrimaryKey": ["StringTypeIndex"], "ParseDates": []},
        TabularDataWithStrings={"PrimaryKey": ["TabularDataIndex"], "ParseDates": []},
        Errors={"PrimaryKey": ["ErrorIndex"], "ParseDates": []},
    )

    zone_weight = ZoneWeight(n=0)

    ep_version: str = Field(
        "9-2-0",
        env="ENERGYPLUS_VERSION",
        description="Latest version of EnergyPlus compatible with archetypal. looks "
        "for ENERGYPLUS_VERSION in os.environ",
    )

    energyplus_location: Optional[DirectoryPath] = Field(
        None,
        env="ENERGYPLUS_LOCATION",
        description="Root directory of the EnergyPlus install.",
    )

    unit_registry: Any = None

    @validator("unit_registry")
    def initialize_units(cls, v):
        if v is not None:
            additional_units = (
                "Dimensionless = dimensionless = Fraction = fraction",
                "@alias degC = Temperature = temperature",
            )
            for unit in additional_units:
                v.define(unit)
        return v


settings = Settings()
settings.unit_registry = unit_registry

# After settings are loaded, import other modules
from .idfclass import IDF
from .eplus_interface.version import EnergyPlusVersion
from .umi_template import UmiTemplateLibrary
from .utils import config, clear_cache, parallel_process
from .umi_template import BuildingTemplate


try:
    __version__ = get_distribution("archetypal").version
except DistributionNotFound:
    # package is not installed
    __version__ = "0.0.0"  # should happen only if package is copied, not installed.
else:
    # warn if a newer version of archetypal is available
    from outdated import warn_if_outdated
    from .eplus_interface.version import warn_if_not_compatible
finally:
    # warn if energyplus not installed or incompatible
    from .eplus_interface.version import warn_if_not_compatible

    warn_if_not_compatible()



"""archetypal OpaqueMaterial."""

import collections

from eppy.bunch_subclass import EpBunch
from validator_collection import validators

from archetypal.template.materials import GasMaterial
from archetypal.template.materials.material_base import MaterialBase
from archetypal.utils import log, signif


class OpaqueMaterial(MaterialBase):
    """Use this component to create a custom opaque material.

    .. image:: ../images/template/materials-opaque.png
    """

    _CREATED_OBJECTS = []

    _ROUGHNESS_TYPES = (
        "VeryRough",
        "Rough",
        "MediumRough",
        "MediumSmooth",
        "Smooth",
        "VerySmooth",
    )

    __slots__ = (
        "_roughness",
        "_solar_absorptance",
        "_specific_heat",
        "_thermal_emittance",
        "_visible_absorptance",
        "_moisture_diffusion_resistance",
        "_conductivity",
        "_density",
        "_key",
    )

    def __init__(
        self,
        Name,
        Conductivity,
        SpecificHeat,
        SolarAbsorptance=0.7,
        ThermalEmittance=0.9,
        VisibleAbsorptance=0.7,
        Roughness="Rough",
        Cost=0,
        Density=1,
        MoistureDiffusionResistance=50,
        EmbodiedCarbon=0.45,
        EmbodiedEnergy=0,
        TransportCarbon=0,
        TransportDistance=0,
        TransportEnergy=0,
        SubstitutionRatePattern=None,
        SubstitutionTimestep=20,
        **kwargs,
    ):
        """Initialize an opaque material.

        Args:
            Name (str): The name of the material.
            Conductivity (float): A number representing the conductivity of the
                material in W/m-K. This is essentially the heat flow in Watts
                across one meter thick of the material when the temperature
                difference on either side is 1 Kelvin. Modeling layers with
                conductivity higher than 5.0 W/(m-K) is not recommended.
            SpecificHeat (float): A number representing the specific heat
                capacity of the material in J/kg-K. This is essentially the
                number of joules needed to raise one kg of the material by 1
                degree Kelvin. Only values of specific heat of 100 or larger are
                allowed. Typical ranges are from 800 to 2000 J/(kg-K).
            SolarAbsorptance (float): An number between 0 and 1 that represents
                the absorptance of solar radiation by the material. The default
                is set to 0.7, which is common for most non-metallic materials.
            ThermalEmittance (float): An number between 0 and 1 that represents
                the thermal absorptance of the material. The default is set to
                0.9, which is common for most non-metallic materials. For long
                wavelength radiant exchange, thermal emissivity and thermal
                emittance are equal to thermal absorptance.
            VisibleAbsorptance (float): An number between 0 and 1 that
                represents the absorptance of visible light by the material.
                The default is set to 0.7, which is common for most non-metallic
                materials.
            Roughness (str): A text value that indicated the roughness of your
                material. This can be either "VeryRough", "Rough",
                "MediumRough", "MediumSmooth", "Smooth", and "VerySmooth". The
                default is set to "Rough".
            Density (float): A number representing the density of the material
                in kg/m3. This is essentially the mass of one cubic meter of the
                material.
            MoistureDiffusionResistance (float): the factor by which the vapor
                diffusion in the material is impeded, as compared to diffusion in
                stagnant air [%].
            **kwargs: keywords passed to parent constructors.
        """
        super(OpaqueMaterial, self).__init__(
            Name,
            Cost=Cost,
            EmbodiedCarbon=EmbodiedCarbon,
            EmbodiedEnergy=EmbodiedEnergy,
            SubstitutionTimestep=SubstitutionTimestep,
            TransportCarbon=TransportCarbon,
            TransportDistance=TransportDistance,
            TransportEnergy=TransportEnergy,
            SubstitutionRatePattern=SubstitutionRatePattern,
            **kwargs,
        )

        self.Conductivity = Conductivity
        self.Density = Density
        self.Roughness = Roughness
        self.SolarAbsorptance = SolarAbsorptance
        self.SpecificHeat = SpecificHeat
        self.ThermalEmittance = ThermalEmittance
        self.VisibleAbsorptance = VisibleAbsorptance
        self.MoistureDiffusionResistance = MoistureDiffusionResistance

        self._key: str = kwargs.get("_key", "")
        # TODO: replace when NoMass and AirGap when properly is supported

        # Only at the end append self to _CREATED_OBJECTS
        self._CREATED_OBJECTS.append(self)

    @property
    def Conductivity(self):
        """Get or set the conductivity of the material [W/m-K]."""
        return self._conductivity

    @Conductivity.setter
    def Conductivity(self, value):
        value = validators.float(value, minimum=0)
        self._conductivity = signif(value)

    @property
    def Density(self):
        """Get or set the density of the material [J/kg-K]."""
        return self._density

    @Density.setter
    def Density(self, value):
        value = validators.float(value, minimum=0)
        self._density = signif(value)

    @property
    def Roughness(self):
        """Get or set the roughness of the material.

        Hint:
            choices are: "VeryRough", "Rough", "MediumRough", "MediumSmooth", "Smooth",
            "VerySmooth".
        """
        return self._roughness

    @Roughness.setter
    def Roughness(self, value):
        assert value in self._ROUGHNESS_TYPES, (
            f"Invalid value '{value}' for material roughness. Roughness must be one "
            f"of the following:\n{self._ROUGHNESS_TYPES}"
        )
        self._roughness = value

    @property
    def SolarAbsorptance(self):
        """Get or set the solar absorptance of the material [-]."""
        return self._solar_absorptance

    @SolarAbsorptance.setter
    def SolarAbsorptance(self, value):
        if value == "" or value is None:
            value = 0.7
        self._solar_absorptance = validators.float(
            value, minimum=0, maximum=1, allow_empty=True
        )

    @property
    def SpecificHeat(self):
        """Get or set the specific heat of the material [J/(kg-K)]."""
        return self._specific_heat

    @SpecificHeat.setter
    def SpecificHeat(self, value):
        value = validators.float(value, minimum=100)
        self._specific_heat = signif(value)

    @property
    def ThermalEmittance(self):
        """Get or set the thermal emittance of the material [-]."""
        return self._thermal_emittance

    @ThermalEmittance.setter
    def ThermalEmittance(self, value):
        if value == "" or value is None:
            value = 0.9
        self._thermal_emittance = validators.float(
            value, minimum=0, maximum=1, allow_empty=True
        )

    @property
    def VisibleAbsorptance(self):
        """Get or set the visible absorptance of the material [-]."""
        return self._visible_absorptance

    @VisibleAbsorptance.setter
    def VisibleAbsorptance(self, value):
        if value == "" or value is None or value is None:
            value = 0.7
        self._visible_absorptance = validators.float(
            value, minimum=0, maximum=1, allow_empty=True
        )

    @property
    def MoistureDiffusionResistance(self):
        """Get or set the vapor resistance factor of the material [%]."""
        return self._moisture_diffusion_resistance

    @MoistureDiffusionResistance.setter
    def MoistureDiffusionResistance(self, value):
        self._moisture_diffusion_resistance = validators.float(value, minimum=0)

    @classmethod
    def generic(cls, **kwargs):
        """Return a generic material based on properties of plaster board.

        Args:
            **kwargs: keywords passed to UmiBase constructor.
        """
        return cls(
            Conductivity=0.16,
            SpecificHeat=1090,
            Density=800,
            Name="GP01 GYPSUM",
            Roughness="Smooth",
            SolarAbsorptance=0.7,
            ThermalEmittance=0.9,
            VisibleAbsorptance=0.5,
            DataSource="ASHRAE 90.1-2007",
            MoistureDiffusionResistance=8.3,
            **kwargs,
        )

    def combine(self, other, weights=None, allow_duplicates=False):
        """Combine two OpaqueMaterial objects.

        Args:
            weights (list-like, optional): A list-like object of len 2. If None,
                the density of the OpaqueMaterial of each objects is used as
                a weighting factor.
            other (OpaqueMaterial): The other OpaqueMaterial object the
                combine with.

        Returns:
            OpaqueMaterial: A new combined object made of self + other.
        """
        # Check if other is the same type as self
        if not isinstance(other, self.__class__):
            msg = "Cannot combine %s with %s" % (
                self.__class__.__name__,
                other.__class__.__name__,
            )
            raise NotImplementedError(msg)

        # Check if other is not the same as self
        if self == other:
            return self

        if not weights:
            log(
                'using OpaqueMaterial density as weighting factor in "{}" '
                "combine.".format(self.__class__.__name__)
            )
            weights = [self.Density, other.Density]

        meta = self._get_predecessors_meta(other)
        new_obj = OpaqueMaterial(
            **meta,
            Conductivity=self.float_mean(other, "Conductivity", weights),
            Roughness=self._str_mean(other, attr="Roughness", append=False),
            SolarAbsorptance=self.float_mean(other, "SolarAbsorptance", weights),
            SpecificHeat=self.float_mean(other, "SpecificHeat"),
            ThermalEmittance=self.float_mean(other, "ThermalEmittance", weights),
            VisibleAbsorptance=self.float_mean(other, "VisibleAbsorptance", weights),
            TransportCarbon=self.float_mean(other, "TransportCarbon", weights),
            TransportDistance=self.float_mean(other, "TransportDistance", weights),
            TransportEnergy=self.float_mean(other, "TransportEnergy", weights),
            SubstitutionRatePattern=self.float_mean(
                other, "SubstitutionRatePattern", weights=None
            ),
            SubstitutionTimestep=self.float_mean(
                other, "SubstitutionTimestep", weights
            ),
            Cost=self.float_mean(other, "Cost", weights),
            Density=self.float_mean(other, "Density", weights),
            EmbodiedCarbon=self.float_mean(other, "EmbodiedCarbon", weights),
            EmbodiedEnergy=self.float_mean(other, "EmbodiedEnergy", weights),
            MoistureDiffusionResistance=self.float_mean(
                other, "MoistureDiffusionResistance", weights
            ),
        )
        new_obj.predecessors.update(self.predecessors + other.predecessors)
        return new_obj

    def to_ref(self):
        """Return a ref pointer to self."""
        pass

    def to_dict(self):
        """Return OpaqueMaterial dictionary representation."""
        self.validate()  # Validate object before trying to get json format

        data_dict = collections.OrderedDict()

        data_dict["$id"] = str(self.id)
        data_dict["MoistureDiffusionResistance"] = self.MoistureDiffusionResistance
        data_dict["Roughness"] = self.Roughness
        data_dict["SolarAbsorptance"] = self.SolarAbsorptance
        data_dict["SpecificHeat"] = self.SpecificHeat
        data_dict["ThermalEmittance"] = self.ThermalEmittance
        data_dict["VisibleAbsorptance"] = self.VisibleAbsorptance
        data_dict["Conductivity"] = self.Conductivity
        data_dict["Cost"] = self.Cost
        data_dict["Density"] = self.Density
        data_dict["EmbodiedCarbon"] = self.EmbodiedCarbon
        data_dict["EmbodiedEnergy"] = self.EmbodiedEnergy
        data_dict["SubstitutionRatePattern"] = self.SubstitutionRatePattern
        data_dict["SubstitutionTimestep"] = self.SubstitutionTimestep
        data_dict["TransportCarbon"] = self.TransportCarbon
        data_dict["TransportDistance"] = self.TransportDistance
        data_dict["TransportEnergy"] = self.TransportEnergy
        data_dict["Category"] = self.Category
        data_dict["Comments"] = validators.string(self.Comments, allow_empty=True)
        data_dict["DataSource"] = self.DataSource
        data_dict["Name"] = self.Name

        return data_dict

    @classmethod
    def from_dict(cls, data, **kwargs):
        """Create an OpaqueMaterial from a dictionary.

        Args:
            data (dict): The python dictionary.
            **kwargs: keywords passed to MaterialBase constructor.

        .. code-block:: python

            {
                "$id": "1",
                "MoistureDiffusionResistance": 50.0,
                "Roughness": "Rough",
                "SolarAbsorptance": 0.7,
                "SpecificHeat": 920.0,
                "ThermalEmittance": 0.9,
                "VisibleAbsorptance": 0.7,
                "Conductivity": 0.85,
                "Cost": 0.0,
                "Density": 2000,
                "EmbodiedCarbon": 0.45,
                "EmbodiedEnergy": 0.0,
                "SubstitutionRatePattern": [
                 1.0
                ],
                "SubstitutionTimestep": 20.0,
                "TransportCarbon": 0.0,
                "TransportDistance": 0.0,
                "TransportEnergy": 0.0,
                "Category": "Uncategorized",
                "Comments": "",
                "DataSource": null,
                "Name": "Concrete"
            }
        """
        _id = data.pop("$id")
        return cls(id=_id, **data, **kwargs)

    @classmethod
    def from_epbunch(cls, epbunch, **kwargs):
        """Create an OpaqueMaterial from an EpBunch.

        Note that "Material", "Material:NoMAss" and "Material:AirGap" objects are
        supported.

        Hint:
            (From EnergyPlus Manual): When a user enters such a “no mass”
            material into EnergyPlus, internally the properties of this layer
            are converted to approximate the properties of air (density,
            specific heat, and conductivity) with the thickness adjusted to
            maintain the user’s desired R-Value. This allowed such layers to be
            handled internally in the same way as other layers without any
            additional changes to the code. This solution was deemed accurate
            enough as air has very little thermal mass and it made the coding of
            the state space method simpler.

            For Material:AirGap, a similar strategy is used, with the
            exception that solar properties (solar and visible absorptance and
            emittance) are assumed null.

        Args:
            epbunch (EpBunch): EP-Construction object
            **kwargs:
        """
        if epbunch.key.upper() == "MATERIAL":
            return cls(
                Conductivity=epbunch.Conductivity,
                Density=epbunch.Density,
                Roughness=epbunch.Roughness,
                SolarAbsorptance=epbunch.Solar_Absorptance,
                SpecificHeat=epbunch.Specific_Heat,
                ThermalEmittance=epbunch.Thermal_Absorptance,
                VisibleAbsorptance=epbunch.Visible_Absorptance,
                Name=epbunch.Name,
                **kwargs,
            )
        elif epbunch.key.upper() == "MATERIAL:NOMASS":
            # Assume properties of air.
            return cls(
                Conductivity=0.02436,  # W/mK, dry air at 0 °C and 100 kPa
                Density=1.2754,  # dry air at 0 °C and 100 kPa.
                Roughness=epbunch.Roughness,
                SolarAbsorptance=epbunch.Solar_Absorptance,
                SpecificHeat=100.5,  # J/kg-K, dry air at 0 °C and 100 kPa
                ThermalEmittance=epbunch.Thermal_Absorptance,
                VisibleAbsorptance=epbunch.Visible_Absorptance,
                Name=epbunch.Name,
                _key=epbunch.key.upper(),
                **kwargs,
            )
        elif epbunch.key.upper() == "MATERIAL:AIRGAP":
            gas_prop = {
                obj.Name.upper(): obj.mapping()
                for obj in [GasMaterial(gas_name) for gas_name in GasMaterial._GASTYPES]
            }
            for gasname, properties in gas_prop.items():
                if gasname.lower() in epbunch.Name.lower():
                    thickness = properties["Conductivity"] * epbunch.Thermal_Resistance
                    properties.pop("Name")
                    return cls(
                        Name=epbunch.Name,
                        Thickness=thickness,
                        SpecificHeat=100.5,
                        _key=epbunch.key.upper(),
                        **properties,
                    )
                else:
                    thickness = (
                        gas_prop["AIR"]["Conductivity"] * epbunch.Thermal_Resistance
                    )
                    properties.pop("Name")
                    return cls(
                        Name=epbunch.Name,
                        Thickness=thickness,
                        SpecificHeat=100.5,
                        _key=epbunch.key.upper(),
                        **gas_prop["AIR"],
                    )
        else:
            raise NotImplementedError(
                "Material '{}' of type '{}' is not yet "
                "supported. Please contact package "
                "authors".format(epbunch.Name, epbunch.key)
            )

    def to_epbunch(self, idf, thickness) -> EpBunch:
        """Convert self to an EpBunch given an idf model and a thickness.

        Args:
            idf (IDF): An IDF model.
            thickness (float): the thickness of the material.

        .. code-block:: python

            MATERIAL,
                ,                         !- Name
                ,                         !- Roughness
                ,                         !- Thickness
                ,                         !- Conductivity
                ,                         !- Density
                ,                         !- Specific Heat
                0.9,                      !- Thermal Absorptance
                0.7,                      !- Solar Absorptance
                0.7;                      !- Visible Absorptance

        Returns:
            EpBunch: The EpBunch object added to the idf model.
        """
        if self._key == "MATERIAL:NOMASS":
            # Special case for Material:NoMass
            return idf.newidfobject(
                "MATERIAL:NOMASS",
                Name=self.Name,
                Roughness=self.Roughness,
                Thermal_Resistance=thickness / self.Conductivity,
                Thermal_Absorptance=self.ThermalEmittance,
                Solar_Absorptance=self.SolarAbsorptance,
                Visible_Absorptance=self.VisibleAbsorptance,
            )
        elif self._key == "MATERIAL:AIRGAP":
            return idf.newidfobject(
                "MATERIAL:AIRGAP",
                Name=self.Name,
                Thermal_Resistance=thickness / self.Conductivity,
            )
        else:
            return idf.newidfobject(
                "MATERIAL",
                Name=self.Name,
                Roughness=self.Roughness,
                Thickness=thickness,
                Conductivity=self.Conductivity,
                Density=self.Density,
                Specific_Heat=self.SpecificHeat,
                Thermal_Absorptance=self.ThermalEmittance,
                Solar_Absorptance=self.SolarAbsorptance,
                Visible_Absorptance=self.VisibleAbsorptance,
            )

    def validate(self):
        """Validate object and fill in missing values.

        Hint:
            Some OpaqueMaterial don't have a default value, therefore an empty string
            is parsed. This breaks the UmiTemplate Editor, therefore we set a value
            on these attributes (if necessary) in this validation step.
        """
        if getattr(self, "SolarAbsorptance") == "":
            setattr(self, "SolarAbsorptance", 0.7)
        if getattr(self, "ThermalEmittance") == "":
            setattr(self, "ThermalEmittance", 0.9)
        if getattr(self, "VisibleAbsorptance") == "":
            setattr(self, "VisibleAbsorptance", 0.7)
        return self

    def mapping(self, validate=False):
        """Get a dict based on the object properties, useful for dict repr.

        Args:
            validate (bool): If True, try to validate object before returning the
                mapping.
        """
        if validate:
            self.validate()

        return dict(
            MoistureDiffusionResistance=self.MoistureDiffusionResistance,
            Roughness=self.Roughness,
            SolarAbsorptance=self.SolarAbsorptance,
            SpecificHeat=self.SpecificHeat,
            ThermalEmittance=self.ThermalEmittance,
            VisibleAbsorptance=self.VisibleAbsorptance,
            Conductivity=self.Conductivity,
            Cost=self.Cost,
            Density=self.Density,
            EmbodiedCarbon=self.EmbodiedCarbon,
            EmbodiedEnergy=self.EmbodiedEnergy,
            SubstitutionRatePattern=self.SubstitutionRatePattern,
            SubstitutionTimestep=self.SubstitutionTimestep,
            TransportCarbon=self.TransportCarbon,
            TransportDistance=self.TransportDistance,
            TransportEnergy=self.TransportEnergy,
            Category=self.Category,
            Comments=self.Comments,
            DataSource=self.DataSource,
            Name=self.Name,
        )

    def duplicate(self):
        """Get copy of self."""
        return self.__copy__()

    def __add__(self, other):
        """Overload + to implement self.combine.

        Args:
            other (OpaqueMaterial):
        """
        return self.combine(other)

    def __hash__(self):
        """Return the hash value of self."""
        return hash(self.id)

    def __eq__(self, other):
        """Assert self is equivalent to other."""
        if not isinstance(other, OpaqueMaterial):
            return NotImplemented
        else:
            return self.__key__() == other.__key__()

    def __key__(self):
        """Get a tuple of attributes. Useful for hashing and comparing."""
        return (
            self.Conductivity,
            self.SpecificHeat,
            self.SolarAbsorptance,
            self.ThermalEmittance,
            self.VisibleAbsorptance,
            self.Roughness,
            self.Cost,
            self.Density,
            self.MoistureDiffusionResistance,
            self.EmbodiedCarbon,
            self.EmbodiedEnergy,
            self.TransportCarbon,
            self.TransportDistance,
            self.TransportEnergy,
            self.SubstitutionRatePattern,
            self.SubstitutionTimestep,
        )

    def __copy__(self):
        """Create a copy of self."""
        new_om = self.__class__(**self.mapping())
        return new_om



import pandas as pd
from energy_pandas import EnergyDataFrame


class LoadBalance:
    def __init__(
        self,
        cooling,
        heating,
        lighting,
        electric_equip,
        gas_equip,
        how_water,
        people_gain,
        solar_gain,
        infiltration,
        mech_vent,
        nat_vent,
        face_energy_flow,
        units="J",
        use_all_solar=True,
    ):
        self.cooling = cooling
        self.heating = heating
        self.lighting = lighting
        self.electric_equip = electric_equip
        self.gas_equip = gas_equip
        self.hot_water = how_water
        self.people_gain = people_gain
        self.solar_gain = solar_gain
        self.infiltration = infiltration
        self.mech_vent = mech_vent
        self.nat_vent = nat_vent
        self.face_energy_flow = face_energy_flow
        self.units = units
        self.use_all_solar = use_all_solar

    @classmethod
    def from_idf(cls, idf):
        assert idf.sql_file.exists()

        # get all of the results relevant for gains and losses
        cooling = idf.variables.OutputVariable.collect_by_output_name(
            idf.outputs.COOLING, reporting_frequency=idf.outputs.reporting_frequency
        )
        heating = idf.variables.OutputVariable.collect_by_output_name(
            idf.outputs.HEATING, reporting_frequency=idf.outputs.reporting_frequency
        )
        lighting = idf.variables.OutputVariable.collect_by_output_name(
            idf.outputs.LIGHTING, reporting_frequency=idf.outputs.reporting_frequency
        )
        people_gain = idf.variables.OutputVariable.collect_by_output_name(
            idf.outputs.PEOPLE_GAIN, reporting_frequency=idf.outputs.reporting_frequency
        )
        solar_gain = idf.variables.OutputVariable.collect_by_output_name(
            idf.outputs.SOLAR_GAIN, reporting_frequency=idf.outputs.reporting_frequency
        )
        infil_gain = idf.variables.OutputVariable.collect_by_output_name(
            idf.outputs.INFIL_GAIN, reporting_frequency=idf.outputs.reporting_frequency
        )
        infil_loss = idf.variables.OutputVariable.collect_by_output_name(
            idf.outputs.INFIL_LOSS, reporting_frequency=idf.outputs.reporting_frequency
        )
        vent_loss = idf.variables.OutputVariable.collect_by_output_name(
            idf.outputs.VENT_LOSS, reporting_frequency=idf.outputs.reporting_frequency
        )
        vent_gain = idf.variables.OutputVariable.collect_by_output_name(
            idf.outputs.VENT_GAIN, reporting_frequency=idf.outputs.reporting_frequency
        )
        nat_vent_gain = idf.variables.OutputVariable.collect_by_output_name(
            idf.outputs.NAT_VENT_GAIN,
            reporting_frequency=idf.outputs.reporting_frequency,
        )
        nat_vent_loss = idf.variables.OutputVariable.collect_by_output_name(
            idf.outputs.NAT_VENT_LOSS,
            reporting_frequency=idf.outputs.reporting_frequency,
        )

        # handle the case that both total elect/gas energy and zone gain are requested
        electric_equip = idf.variables.OutputVariable.collect_by_output_name(
            idf.outputs.ELECTRIC_EQUIP[1],
            reporting_frequency=idf.outputs.reporting_frequency,
        )
        if len(electric_equip) == 0:
            electric_equip = idf.variables.OutputVariable.collect_by_output_name(
                idf.outputs.ELECTRIC_EQUIP,
                reporting_frequency=idf.outputs.reporting_frequency,
            )
        gas_equip = idf.variables.OutputVariable.collect_by_output_name(
            idf.outputs.GAS_EQUIP[1],
            reporting_frequency=idf.outputs.reporting_frequency,
        )
        if len(gas_equip) == 0:
            gas_equip = idf.variables.OutputVariable.collect_by_output_name(
                idf.outputs.GAS_EQUIP,
                reporting_frequency=idf.outputs.reporting_frequency,
            )
        hot_water = idf.variables.OutputVariable.collect_by_output_name(
            idf.outputs.HOT_WATER[1],
            reporting_frequency=idf.outputs.reporting_frequency,
        )
        if len(hot_water) == 0:
            hot_water = idf.variables.OutputVariable.collect_by_output_name(
                idf.outputs.HOT_WATER,
                reporting_frequency=idf.outputs.reporting_frequency,
            )

        # subtract losses from gains
        infiltration = None
        mech_vent = None
        nat_vent = None
        if len(infil_gain) == len(infil_loss):
            infiltration = cls.subtract_loss_from_gain(infil_gain, infil_loss)
        if (
            vent_gain.shape == vent_loss.shape == cooling.shape == heating.shape
            and not vent_gain.empty == vent_loss.empty == cooling.empty == heating.empty
        ):
            mech_vent_loss = cls.subtract_loss_from_gain(heating, vent_loss)
            mech_vent_gain = cls.subtract_loss_from_gain(cooling, vent_gain)
            total_load = cls.subtract_loss_from_gain(mech_vent_gain, mech_vent_loss)
            mech_vent = total_load.copy()
            mech_vent.rename(
                columns=lambda x: str.replace(
                    x, "Zone Ideal Loads Supply Air", "Zone Ideal Loads Ventilation"
                ),
                level="OutputVariable",
                inplace=True,
            )
        if nat_vent_gain.shape == nat_vent_loss.shape:
            nat_vent = cls.subtract_loss_from_gain(nat_vent_gain, nat_vent_loss)

        # get the surface energy flow
        opaque_flow = idf.variables.OutputVariable.collect_by_output_name(
            idf.outputs.OPAQUE_ENERGY_FLOW,
            reporting_frequency=idf.outputs.reporting_frequency,
        )
        window_loss = idf.variables.OutputVariable.collect_by_output_name(
            idf.outputs.WINDOW_LOSS, reporting_frequency=idf.outputs.reporting_frequency
        )
        window_gain = idf.variables.OutputVariable.collect_by_output_name(
            idf.outputs.WINDOW_GAIN, reporting_frequency=idf.outputs.reporting_frequency
        )
        window_flow = []
        if window_gain.shape == window_loss.shape:
            window_flow = cls.subtract_loss_from_gain(window_gain, window_loss)
            window_flow = cls.match_window_to_zone(idf, window_flow)
        face_energy_flow = opaque_flow.add(
            window_flow.sum(level=["Building_Surface_Name"], axis=1).rename(
                columns=str.upper
            ),
            level="Key_Name",
            axis=1,
            fill_value=0,
        )

        bal_obj = cls(
            cooling,
            heating,
            lighting,
            electric_equip,
            gas_equip,
            hot_water,
            people_gain,
            solar_gain,
            infiltration,
            mech_vent,
            nat_vent,
            face_energy_flow,
            "J",
            use_all_solar=True,
        )
        return bal_obj

    @classmethod
    def match_window_to_zone(cls, idf, window_flow):
        """Match the DataFrame of"""
        assert window_flow.columns.names == ["OutputVariable", "Key_Name"]
        window_to_surface_match = pd.DataFrame(
            [
                (
                    window.Name,  # name of the window
                    window.Building_Surface_Name,  # name of the wall this window is on
                    window.get_referenced_object(
                        "Building_Surface_Name"
                    ).Surface_Type,  # surface type (wall, ceiling, floor) this windows is on.
                    window.get_referenced_object(  # get the zone name though the surface name
                        "Building_Surface_Name"
                    ).Zone_Name,
                    window.Multiplier,  # multiplier of this window.
                )
                for window in idf.getsubsurfaces()
            ],
            columns=[
                "Name",
                "Building_Surface_Name",
                "Surface_Type",
                "Zone_Name",
                "Multiplier",
            ],
        ).set_index("Name")
        # Match the subsurface to the surface name and the zone name it belongs to.
        stacked = (
            window_flow.stack()
            .join(
                window_to_surface_match.rename(index=str.upper),
                on="Key_Name",
            )
            .set_index(
                ["Building_Surface_Name", "Surface_Type", "Zone_Name"], append=True
            )
        )
        window_flow = stacked.drop(columns=["Multiplier"]).iloc[:, 0] * pd.to_numeric(
            stacked["Multiplier"]
        )
        window_flow = window_flow.unstack(
            level=["Key_Name", "Building_Surface_Name", "Surface_Type", "Zone_Name"]
        )

        return window_flow  # .groupby("Building_Surface_Name", axis=1).sum()

    @classmethod
    def subtract_loss_from_gain(cls, load_gain, load_loss):
        try:
            columns = load_gain.rename(
                columns=lambda x: str.replace(x, " Gain", ""), level="OutputVariable"
            ).columns
        except KeyError:
            columns = None
        return EnergyDataFrame(
            load_gain.values - load_loss.values,
            columns=columns,
            index=load_gain.index,
        )

    def to_df(self):
        return pd.concat(
            [
                df
                for df in [
                    self.cooling,
                    self.heating,
                    self.lighting,
                    self.electric_equip,
                    self.gas_equip,
                    self.hot_water,
                    self.people_gain,
                    self.solar_gain,
                    self.infiltration,
                    self.mech_vent,
                    self.nat_vent,
                    self.face_energy_flow,
                ]
                if not df.empty
            ],
            axis=1,
            verify_integrity=True,
        )



"""archetypal MaterialBase."""

import numpy as np
from validator_collection import validators

from archetypal.template.umi_base import UmiBase


class MaterialBase(UmiBase):
    """A class used to store data linked with the Life Cycle aspect of materials.

    For more information on the Life Cycle Analysis performed in UMI, see:
    https://umidocs.readthedocs.io/en/latest/docs/life-cycle-introduction.html#life
    -cycle-impact
    """

    __slots__ = (
        "_cost",
        "_embodied_carbon",
        "_embodied_energy",
        "_substitution_timestep",
        "_transport_carbon",
        "_transport_distance",
        "_transport_energy",
        "_substitution_rate_pattern",
        "_density",
    )

    def __init__(
        self,
        Name,
        Cost=0,
        EmbodiedCarbon=0,
        EmbodiedEnergy=0,
        SubstitutionTimestep=100,
        TransportCarbon=0,
        TransportDistance=0,
        TransportEnergy=0,
        SubstitutionRatePattern=None,
        **kwargs,
    ):
        """Initialize a MaterialBase object with parameters.

        Args:
            Name (str): Name of the Material.
            Cost (float): The purchase cost of the material by volume ($/m3).
            EmbodiedCarbon (float): Represents the GHG emissions through the
                lifetime of the product (kgCO2/kg).
            EmbodiedEnergy (float): Represents all fuel consumption ( Typically
                from non-renewable sources) which happened through the lifetime
                of a product (or building), expressed as primary energy (MJ/kg).
            SubstitutionTimestep (float): The duration in years of a period of
                replacement (e.g. There will be interventions in this material
                type every 10 years).
            TransportCarbon (float): The impacts associated with the transport
                by km of distance and kg of material (kgCO2/kg/km).
            TransportDistance (float): The average distance in km from the
                manufacturing site to the building construction site
            TransportEnergy (float): The impacts associated with the transport
                by km of distance and kg of material (MJ/kg/km).
            SubstitutionRatePattern (list-like): A ratio from 0 to 1 which
                defines the amount of the material replaced at the end of each
                period of replacement, :attr:`SubstitutionTimestep` (e.g. Every
                10 years this cladding will be completely replaced with ratio
                1). Notice that you can define different replacement ratios for
                different consecutive periods, introducing them separated by
                commas. For example, if you introduce the series “0.1 , 0.1 , 1”
                after the first 10 years a 10% will be replaced, then after 20
                years another 10%, then after 30 years a 100%, and finally the
                series would start again in year 40.
            **kwargs: Keywords passed to the :class:`UmiBase` class. See
                :class:`UmiBase` for more details.
        """
        super(MaterialBase, self).__init__(Name, **kwargs)
        self.Cost = Cost
        self.EmbodiedCarbon = EmbodiedCarbon
        self.EmbodiedEnergy = EmbodiedEnergy
        self.SubstitutionRatePattern = SubstitutionRatePattern
        self.SubstitutionTimestep = SubstitutionTimestep
        self.TransportCarbon = TransportCarbon
        self.TransportDistance = TransportDistance
        self.TransportEnergy = TransportEnergy

    @property
    def Cost(self):
        """Get or set the cost of the material [$]."""
        return self._cost

    @Cost.setter
    def Cost(self, value):
        self._cost = validators.float(value)

    @property
    def EmbodiedCarbon(self):
        """Get or set the embodied carbon of the material [kgCO2/kg]."""
        return self._embodied_carbon

    @EmbodiedCarbon.setter
    def EmbodiedCarbon(self, value):
        self._embodied_carbon = validators.float(value)

    @property
    def EmbodiedEnergy(self):
        """Get or set the embodied energy of the material [MJ/kg]."""
        return self._embodied_energy

    @EmbodiedEnergy.setter
    def EmbodiedEnergy(self, value):
        self._embodied_energy = validators.float(value)

    @property
    def SubstitutionTimestep(self):
        """Get or set the substitution timestep of the material."""
        return self._substitution_timestep

    @SubstitutionTimestep.setter
    def SubstitutionTimestep(self, value):
        self._substitution_timestep = validators.float(value, minimum=0)

    @property
    def SubstitutionRatePattern(self):
        """Get or set the substitution rate pattern of the material."""
        return self._substitution_rate_pattern

    @SubstitutionRatePattern.setter
    def SubstitutionRatePattern(self, value):
        if value is None:
            value = [1.0]
        elif isinstance(value, np.ndarray):
            value = value.tolist()
        self._substitution_rate_pattern = validators.iterable(value, allow_empty=True)

    @property
    def TransportCarbon(self):
        """Get or set the transportation carbon of the material [kgCO2/kg/km]."""
        return self._transport_carbon

    @TransportCarbon.setter
    def TransportCarbon(self, value):
        self._transport_carbon = validators.float(value, minimum=0)

    @property
    def TransportDistance(self):
        """Get or set the transportation distance of the material [km]."""
        return self._transport_distance

    @TransportDistance.setter
    def TransportDistance(self, value):
        self._transport_distance = validators.float(value, minimum=0)

    @property
    def TransportEnergy(self):
        """Get or set the transporation energy of the material [MJ/kg/km]."""
        return self._transport_energy

    @TransportEnergy.setter
    def TransportEnergy(self, value):
        self._transport_energy = validators.float(value, minimum=0)

    def __hash__(self):
        """Return the hash value of self."""
        return hash(self.id)

    def __eq__(self, other):
        """Assert self is equivalent to other."""
        if not isinstance(other, MaterialBase):
            return NotImplemented
        else:
            return all(
                [
                    self.Cost == other.Cost,
                    self.EmbodiedCarbon == other.EmbodiedCarbon,
                    self.EmbodiedEnergy == other.EmbodiedEnergy,
                    self.SubstitutionTimestep == other.SubstitutionTimestep,
                    self.TransportCarbon == other.TransportCarbon,
                    self.TransportDistance == other.TransportDistance,
                    self.TransportEnergy == other.TransportEnergy,
                    np.array_equal(
                        self.SubstitutionRatePattern, other.SubstitutionRatePattern
                    ),
                    self.Density == other.Density,
                ]
            )

    def validate(self):
        """Validate object and fill in missing values."""
        return self



from sqlite3 import connect

import numpy as np
import pandas as pd
from energy_pandas import EnergyDataFrame
from energy_pandas.units import unit_registry

from archetypal.idfclass.sql import Sql


class EndUseBalance:
    HVAC_MODE = ("Zone Predicted Sensible Load to Setpoint Heat Transfer Rate",)
    HVAC_INPUT_SENSIBLE = (  # not multiplied by zone or group multipliers
        "Zone Air Heat Balance System Air Transfer Rate",
        "Zone Air Heat Balance System Convective Heat Gain Rate",
    )
    HVAC_INPUT_HEATED_SURFACE = (
        "Zone Radiant HVAC Heating Energy",
        "Zone Ventilated Slab Radiant Heating Energy",
    )
    HVAC_INPUT_COOLED_SURFACE = (
        "Zone Radiant HVAC Cooling Energy",
        "Zone Ventilated Slab Radiant Cooling Energy",
    )
    LIGHTING = ("Zone Lights Total Heating Energy",)  # checked
    EQUIP_GAINS = (  # checked
        "Zone Electric Equipment Radiant Heating Energy",
        "Zone Gas Equipment Radiant Heating Energy",
        "Zone Steam Equipment Radiant Heating Energy",
        "Zone Hot Water Equipment Radiant Heating Energy",
        "Zone Other Equipment Radiant Heating Energy",
        "Zone Electric Equipment Convective Heating Energy",
        "Zone Gas Equipment Convective Heating Energy",
        "Zone Steam Equipment Convective Heating Energy",
        "Zone Hot Water Equipment Convective Heating Energy",
        "Zone Other Equipment Convective Heating Energy",
    )
    PEOPLE_GAIN = ("Zone People Total Heating Energy",)  # checked
    SOLAR_GAIN = ("Zone Windows Total Transmitted Solar Radiation Energy",)  # checked
    INFIL_GAIN = (
        "Zone Infiltration Total Heat Gain Energy",  # checked
        "AFN Zone Infiltration Total Heat Gain Energy",
    )
    INFIL_LOSS = (
        "Zone Infiltration Total Heat Loss Energy",  # checked
        "AFN Zone Infiltration Total Heat Loss Energy",
    )
    VENTILATION_LOSS = ("Zone Air System Total Heating Energy",)
    VENTILATION_GAIN = ("Zone Air System Total Cooling Energy",)
    NAT_VENT_GAIN = (
        "Zone Ventilation Total Heat Gain Energy",
        "AFN Zone Ventilation Total Heat Gain Energy",
    )
    NAT_VENT_LOSS = (
        "Zone Ventilation Total Heat Loss Energy",
        "AFN Zone Ventilation Total Heat Loss Energy",
    )
    OPAQUE_ENERGY_FLOW = ("Surface Outside Face Conduction Heat Transfer Energy",)
    OPAQUE_ENERGY_STORAGE = ("Surface Heat Storage Energy",)
    WINDOW_LOSS = ("Zone Windows Total Heat Loss Energy",)  # checked
    WINDOW_GAIN = ("Zone Windows Total Heat Gain Energy",)  # checked
    HRV_LOSS = ("Heat Exchanger Total Cooling Energy",)
    HRV_GAIN = ("Heat Exchanger Total Heating Energy",)
    AIR_SYSTEM = (
        "Air System Heating Coil Total Heating Energy",
        "Air System Cooling Coil Total Cooling Energy",
    )

    def __init__(
        self,
        sql_file,
        cooling,
        heating,
        lighting,
        electric_equip,
        gas_equip,
        how_water,
        people_gain,
        solar_gain,
        infiltration,
        mech_vent,
        nat_vent,
        window_energy_flow,
        opaque_flow,
        opaque_storage,
        window_flow,
        heat_recovery,
        air_system,
        is_cooling,
        is_heating,
        units="J",
        use_all_solar=True,
    ):
        self.sql_file = sql_file
        self.cooling = cooling
        self.heating = heating
        self.lighting = lighting
        self.electric_equip = electric_equip
        self.gas_equip = gas_equip
        self.hot_water = how_water
        self.people_gain = people_gain
        self.solar_gain = solar_gain
        self.infiltration = infiltration
        self.mech_vent = mech_vent
        self.nat_vent = nat_vent
        self.window_energy_flow = window_energy_flow
        self.opaque_flow = opaque_flow
        self.opaque_storage = opaque_storage
        self.window_flow = window_flow
        self.heat_recovery = heat_recovery
        self.air_system = air_system
        self.units = units
        self.use_all_solar = use_all_solar
        self.is_cooling = is_cooling
        self.is_heating = is_heating

    @classmethod
    def from_sql_file(
        cls, sql_file, units="kWh", power_units="kW", outdoor_surfaces_only=True
    ):
        sql = Sql(sql_file)

        _hvac_input = sql.timeseries_by_name(cls.HVAC_INPUT_SENSIBLE).to_units(
            power_units
        )
        _hvac_input_heated_surface = sql.timeseries_by_name(
            cls.HVAC_INPUT_HEATED_SURFACE
        ).to_units(units)
        _hvac_input_cooled_surface = sql.timeseries_by_name(
            cls.HVAC_INPUT_COOLED_SURFACE
        ).to_units(units)
        # convert power to energy assuming the reporting frequency
        freq = pd.infer_freq(_hvac_input.index)
        assert freq == "H", "A reporting frequency other than H is not yet supported."
        freq_to_unit = {"H": "hr"}
        _hvac_input = _hvac_input.apply(
            lambda row: unit_registry.Quantity(
                row.values,
                unit_registry(power_units) * unit_registry(freq_to_unit[freq]),
            )
            .to(units)
            .m
        )

        _hvac_input = pd.concat(
            filter(
                lambda x: not x.empty,
                [
                    _hvac_input,
                    EndUseBalance.subtract_cooled_from_heated_surface(
                        _hvac_input_cooled_surface, _hvac_input_heated_surface
                    ),
                ],
            ),
            axis=1,
            verify_integrity=True,
        )
        mode = sql.timeseries_by_name(cls.HVAC_MODE)  # positive = Heating
        rolling_sign = cls.get_rolling_sign_change(mode).fillna(0)

        # Create both heating and cooling masks
        is_heating = rolling_sign.droplevel(["IndexGroup", "Name"], axis=1) == 1
        is_cooling = rolling_sign.droplevel(["IndexGroup", "Name"], axis=1) == -1

        heating = _hvac_input.mul(is_heating, level="KeyValue", axis=1)
        cooling = _hvac_input.mul(is_cooling, level="KeyValue", axis=1)

        lighting = sql.timeseries_by_name(cls.LIGHTING).to_units(units)
        zone_multipliers = sql.zone_info.set_index("ZoneName")["Multiplier"].rename(
            "KeyValue"
        )
        lighting = cls.apply_multipliers(
            lighting,
            zone_multipliers,
        )
        people_gain = sql.timeseries_by_name(cls.PEOPLE_GAIN).to_units(units)
        people_gain = cls.apply_multipliers(people_gain, zone_multipliers)
        equipment = sql.timeseries_by_name(cls.EQUIP_GAINS).to_units(units)
        equipment = cls.apply_multipliers(equipment, zone_multipliers)
        solar_gain = sql.timeseries_by_name(cls.SOLAR_GAIN).to_units(units)
        solar_gain = cls.apply_multipliers(solar_gain, zone_multipliers)
        infil_gain = sql.timeseries_by_name(cls.INFIL_GAIN).to_units(units)
        infil_gain = cls.apply_multipliers(infil_gain, zone_multipliers)
        infil_loss = sql.timeseries_by_name(cls.INFIL_LOSS).to_units(units)
        infil_loss = cls.apply_multipliers(infil_loss, zone_multipliers)
        vent_loss = sql.timeseries_by_name(cls.VENTILATION_LOSS).to_units(units)
        vent_loss = cls.apply_multipliers(vent_loss, zone_multipliers)
        vent_gain = sql.timeseries_by_name(cls.VENTILATION_GAIN).to_units(units)
        vent_gain = cls.apply_multipliers(vent_gain, zone_multipliers)
        nat_vent_gain = sql.timeseries_by_name(cls.NAT_VENT_GAIN).to_units(units)
        nat_vent_gain = cls.apply_multipliers(nat_vent_gain, zone_multipliers)
        nat_vent_loss = sql.timeseries_by_name(cls.NAT_VENT_LOSS).to_units(units)
        nat_vent_loss = cls.apply_multipliers(nat_vent_loss, zone_multipliers)
        hrv_loss = sql.timeseries_by_name(cls.HRV_LOSS).to_units(units)
        hrv_gain = sql.timeseries_by_name(cls.HRV_GAIN).to_units(units)
        hrv = cls.subtract_loss_from_gain(hrv_gain, hrv_loss, level="KeyValue")
        air_system = sql.timeseries_by_name(cls.AIR_SYSTEM).to_units(units)

        # subtract losses from gains
        infiltration = None
        mech_vent = None
        nat_vent = None
        if len(infil_gain) == len(infil_loss):
            infiltration = cls.subtract_loss_from_gain(
                infil_gain, infil_loss, level="Name"
            )
        if nat_vent_gain.shape == nat_vent_loss.shape:
            nat_vent = cls.subtract_loss_from_gain(
                nat_vent_gain, nat_vent_loss, level="Name"
            )

        # get the surface energy flow
        opaque_flow = sql.timeseries_by_name(cls.OPAQUE_ENERGY_FLOW).to_units(units)
        opaque_storage = sql.timeseries_by_name(cls.OPAQUE_ENERGY_STORAGE).to_units(
            units
        )
        opaque_storage_ = opaque_storage.copy()
        opaque_storage_.columns = opaque_flow.columns
        opaque_flow = -(opaque_flow + opaque_storage_)
        window_loss = sql.timeseries_by_name(cls.WINDOW_LOSS).to_units(units)
        window_loss = cls.apply_multipliers(window_loss, zone_multipliers)
        window_gain = sql.timeseries_by_name(cls.WINDOW_GAIN).to_units(units)
        window_gain = cls.apply_multipliers(window_gain, zone_multipliers)
        window_flow = cls.subtract_loss_from_gain(
            window_gain, window_loss, level="Name"
        )
        window_flow = cls.subtract_solar_from_window_net(
            window_flow, solar_gain, level="KeyValue"
        )

        opaque_flow = cls.match_opaque_surface_to_zone(
            sql.surfaces_table, opaque_flow, sql.zone_info
        )
        opaque_storage = cls.match_opaque_surface_to_zone(
            sql.surfaces_table, opaque_storage, sql.zone_info
        )
        if outdoor_surfaces_only:
            # inside surfaces are identified by ExtBoundCond > 0
            inside_surfaces = sql.surfaces_table[lambda x: x["ExtBoundCond"] > 0][
                "SurfaceName"
            ].values.tolist()

            # drop inside surfaces
            opaque_flow = opaque_flow.drop(
                inside_surfaces, level="KeyValue", axis=1, errors="ignore"
            )
            opaque_storage = opaque_storage.drop(
                inside_surfaces, level="KeyValue", axis=1, errors="ignore"
            )
        window_energy_flow = window_flow

        bal_obj = cls(
            sql_file,
            cooling,
            heating,
            lighting,
            equipment,
            None,
            None,
            people_gain,
            solar_gain,
            infiltration,
            mech_vent,
            nat_vent,
            window_energy_flow,
            opaque_flow,
            opaque_storage,
            window_flow,
            hrv,
            air_system,
            is_cooling,
            is_heating,
            units,
            use_all_solar=True,
        )
        return bal_obj

    @classmethod
    def apply_multipliers(cls, data, idf):
        from archetypal import IDF

        if isinstance(idf, IDF):
            multipliers = (
                pd.Series(
                    {
                        zone.Name.upper(): zone.Multiplier
                        for zone in idf.idfobjects["ZONE"]
                    },
                    name="Key_Name",
                )
                .replace({"": 1})
                .fillna(1)
            )
            key = "OutputVariable"
        elif isinstance(idf, pd.Series):
            multipliers = idf
            key = "KeyValue"
        else:
            raise ValueError
        return data.mul(multipliers, level=key, axis=1)

    @classmethod
    def subtract_cooled_from_heated_surface(
        cls, _hvac_input_cooled_surface, _hvac_input_heated_surface
    ):
        if _hvac_input_cooled_surface.empty:
            return _hvac_input_cooled_surface
        try:
            columns = _hvac_input_heated_surface.rename(
                columns=lambda x: str.replace(x, " Heating", ""), level="OutputVariable"
            ).columns
        except KeyError:
            columns = None
        return EnergyDataFrame(
            (
                _hvac_input_heated_surface.sum(level="KeyValue", axis=1)
                - _hvac_input_cooled_surface.sum(level="KeyValue", axis=1)
            ).values,
            columns=columns,
            index=_hvac_input_heated_surface.index,
        )

    @classmethod
    def get_rolling_sign_change(cls, data: pd.DataFrame):
        # create a sign series where -1 is negative and 0 or 1 is positive
        sign = (
            np.sign(data)
            .replace({0: np.NaN})
            .fillna(method="bfill")
            .fillna(method="ffill")
        )
        # when does a change of sign occurs?
        sign_switch = sign != sign.shift(-1)
        # From sign, keep when the sign switches and fill with the previous values
        # (back fill). The final forward fill is to fill the last few timesteps of the
        # series which might be NaNs.
        rolling_sign = sign[sign_switch].fillna(method="bfill").fillna(method="ffill")
        return rolling_sign

    @classmethod
    def match_window_to_zone(cls, idf, window_flow):
        """Match window surfaces with their wall and zone.

        Adds the following properties to the `window_flow` DataFrame as a MultiIndex level with names:
            * Building_Surface_Name
            * Surface_Type
            * Zone_Name
            * Multiplier
        """
        # Todo: Check if Zone Multiplier needs to be added.
        assert window_flow.columns.names == ["OutputVariable", "Key_Name"]
        window_to_surface_match = pd.DataFrame(
            [
                (
                    window.Name.upper(),  # name of the window
                    window.Building_Surface_Name.upper(),  # name of the wall this window is on
                    window.get_referenced_object(
                        "Building_Surface_Name"
                    ).Surface_Type.title(),  # surface type (wall, ceiling, floor) this windows is on.
                    window.get_referenced_object(  # get the zone name though the surface name
                        "Building_Surface_Name"
                    ).Zone_Name.upper(),
                    float(window.Multiplier)
                    if window.Multiplier != ""
                    else 1,  # multiplier of this window.
                )
                for window in idf.getsubsurfaces()
            ],
            columns=[
                "Name",
                "Building_Surface_Name",
                "Surface_Type",
                "Zone_Name",
                "Multiplier",
            ],
        ).set_index("Name")
        # Match the subsurface to the surface name and the zone name it belongs to.
        stacked = (
            window_flow.stack()
            .join(
                window_to_surface_match.rename(index=str.upper),
                on="Key_Name",
            )
            .set_index(
                ["Building_Surface_Name", "Surface_Type", "Zone_Name"], append=True
            )
        )
        window_flow = (
            stacked.drop(columns=["Multiplier"]).iloc[:, 0] * stacked["Multiplier"]
        )
        window_flow = window_flow.unstack(
            level=["Key_Name", "Building_Surface_Name", "Surface_Type", "Zone_Name"]
        )

        return window_flow  # .groupby("Building_Surface_Name", axis=1).sum()

    @classmethod
    def match_opaque_surface_to_zone(cls, surface_table, opaque_flow, zone_info):
        """Match opaque surfaces with their zone.

        Multiplies the surface heat flow by the zone multiplier.

        Adds the following properties to the `opaque_flow` DataFrame as a MultiIndex level with names:
            * Surface_Type
            * Outside_Boundary_Condition
            * Zone_Name
        """
        surface_with_idx = surface_table.join(
            zone_info["ZoneName"], on="ZoneIndex"
        ).reset_index()
        opaque_flow.columns = pd.MultiIndex.from_frame(
            opaque_flow.columns.to_frame(index=False).join(
                surface_with_idx.reset_index()
                .set_index("SurfaceName")[
                    ["ClassName", "ExtBoundCond", "ZoneName", "ZoneIndex"]
                ]
                .rename(
                    {
                        "ClassName": "Surface_Type",
                        "ExtBoundCond": "Outside_Boundary_Condition",
                        "ZoneName": "Zone_Name",
                    },
                    axis=1,
                ),
                on="KeyValue",
            )
        )
        opaque_flow.mul(zone_info["Multiplier"], level="ZoneIndex")
        opaque_flow = opaque_flow.droplevel("ZoneIndex", axis=1)
        return opaque_flow

    @classmethod
    def subtract_loss_from_gain(cls, load_gain, load_loss, level="OutputVariable"):
        try:
            columns = load_gain.rename(
                columns=lambda x: str.replace(x, " Gain", ""), level=level
            ).columns
        except KeyError:
            columns = None
        return EnergyDataFrame(
            load_gain.values - load_loss.values,
            columns=columns,
            index=load_gain.index,
        )

    @classmethod
    def subtract_solar_from_window_net(cls, window_flow, solar_gain, level="Key_Name"):
        columns = window_flow.columns
        return EnergyDataFrame(
            (window_flow.sum(level=level, axis=1) - solar_gain.sum(level=level, axis=1))
            .loc[:, list(columns.get_level_values(level))]
            .values,
            columns=columns,
            index=window_flow.index,
        )

    @classmethod
    def subtract_vent_from_system(cls, system, vent, level="Key_Name"):
        columns = vent.columns
        return EnergyDataFrame(
            system.sum(level=level, axis=1).values
            - vent.sum(level=level, axis=1).values,
            columns=columns,
            index=system.index,
        )

    def separate_gains_and_losses(self, component, level="Key_Name") -> EnergyDataFrame:
        """Separate gains from losses when cooling and heating occurs for the component.

        Args:
            component (str):
            level (str or list):

        Returns:

        """
        assert (
            component in self.__dict__.keys()
        ), f"{component} is not a valid attribute of EndUseBalance."
        component_df = getattr(self, component)
        assert not component_df.empty, "Expected a component that is not empty."
        print(component)

        c_df = getattr(self, component)

        # concatenate the Periods
        inter = pd.concat(
            [
                c_df.mul(self.is_cooling.rename_axis(level, axis=1), level=level),
                c_df.mul(self.is_heating.rename_axis(level, axis=1), level=level),
            ],
            keys=["Cooling Periods", "Heating Periods"],
            names=["Period"],
        )

        # mask when values are positive (gain)
        positive_mask = inter >= 0

        # concatenate the Gain/Loss
        final = pd.concat(
            [
                inter[positive_mask].reindex(inter.index),
                inter[~positive_mask].reindex(inter.index),
            ],
            keys=["Heat Gain", "Heat Loss"],
            names=["Gain/Loss"],
        ).unstack(["Period", "Gain/Loss"])
        final.sort_index(axis=1, inplace=True)
        return final

    def to_df(self, separate_gains_and_losses=False, level="KeyValue"):
        """Summarize components into a DataFrame."""
        if separate_gains_and_losses:
            summary_by_component = {}
            levels = ["Component", "Zone_Name", "Period", "Gain/Loss"]
            for component in [
                "cooling",
                "heating",
                "lighting",
                "electric_equip",
                "people_gain",
                "solar_gain",
                "infiltration",
                "window_energy_flow",
                "nat_vent",
            ]:
                if not getattr(self, component).empty:
                    summary_by_component[component] = (
                        self.separate_gains_and_losses(
                            component,
                            level=level,
                        )
                        .groupby(level=["KeyValue", "Period", "Gain/Loss"], axis=1)
                        .sum()
                        .sort_index(axis=1)
                    )
            for (surface_type), data in self.separate_gains_and_losses(
                "opaque_flow", level="Zone_Name"
            ).groupby(level=["Surface_Type"], axis=1):
                summary_by_component[surface_type] = data.sum(
                    level=["Zone_Name", "Period", "Gain/Loss"], axis=1
                ).sort_index(axis=1)

        else:
            summary_by_component = {}
            for component in [
                "cooling",
                "heating",
                "lighting",
                "electric_equip",
                "people_gain",
                "solar_gain",
                "infiltration",
                "window_energy_flow",
                "nat_vent",
            ]:
                component_df = getattr(self, component)
                if not component_df.empty:
                    summary_by_component[component] = component_df.sum(
                        level=level, axis=1
                    ).sort_index(axis=1)
            for (zone_name, surface_type), data in self.opaque_flow.groupby(
                level=["Zone_Name", "Surface_Type"], axis=1
            ):
                summary_by_component[surface_type] = data.sum(
                    level="Zone_Name", axis=1
                ).sort_index(axis=1)
            levels = ["Component", "Zone_Name"]

        # Add contribution of heating/cooling outside air, if any
        if not self.air_system.empty:
            summary_by_component["air_system_heating"] = (
                self.air_system.xs(
                    "Air System Heating Coil Total Heating Energy", level="Name", axis=1
                )
                .assign(**{"Period": "Heating Periods", "Gain/Loss": "Heat Loss"})
                .set_index(["Period", "Gain/Loss"], append=True)
                .unstack(["Period", "Gain/Loss"])
                .droplevel("IndexGroup", axis=1)
            )
            summary_by_component["air_system_cooling"] = (
                self.air_system.xs(
                    "Air System Cooling Coil Total Cooling Energy", level="Name", axis=1
                )
                .assign(**{"Period": "Cooling Periods", "Gain/Loss": "Heat Gain"})
                .set_index(["Period", "Gain/Loss"], append=True)
                .unstack(["Period", "Gain/Loss"])
                .droplevel("IndexGroup", axis=1)
            )
        return pd.concat(
            summary_by_component, axis=1, verify_integrity=True, names=levels
        )

    def component_summary(self) -> EnergyDataFrame:
        """Return a DataFrame of components summarized annually."""
        sum_opaque_flow = (
            self.separate_gains_and_losses("opaque_flow", "Zone_Name")
            .sum()
            .sum(level=["Period", "Gain/Loss"])
        )
        sum_window_flow = (
            self.separate_gains_and_losses("window_flow", "Zone_Name")
            .sum()
            .sum(level=["Period", "Gain/Loss"])
        )
        sum_solar_gain = (
            self.separate_gains_and_losses("solar_gain")
            .sum()
            .sum(level=["Period", "Gain/Loss"])
        )
        sum_lighting = (
            self.separate_gains_and_losses("lighting")
            .sum()
            .sum(level=["Period", "Gain/Loss"])
        )
        sum_infiltration = (
            self.separate_gains_and_losses("infiltration")
            .sum()
            .sum(level=["Period", "Gain/Loss"])
        )
        sum_people_gain = (
            self.separate_gains_and_losses("people_gain")
            .sum()
            .sum(level=["Period", "Gain/Loss"])
        )

        df = pd.concat(
            [
                sum_opaque_flow,
                sum_window_flow,
                sum_solar_gain,
                sum_lighting,
                sum_infiltration,
                sum_people_gain,
            ],
            keys=[
                "Opaque Conduction",
                "Window Conduction",
                "Window Solar Gains",
                "Lighting",
                "Infiltration",
                "Occupants (Sensible + Latent)",
            ],
        )

        return df.unstack(level=["Period", "Gain/Loss"])

    def to_sankey(self, path_or_buf):
        system_data = self.to_df(separate_gains_and_losses=True)
        annual_system_data = system_data.sum().sum(
            level=["Component", "Period", "Gain/Loss"]
        )
        annual_system_data.rename(
            {
                "people_gain": "Occupants",
                "solar_gain": "Passive Solar",
                "electric_equip": "Equipment",
                "lighting": "Lighting",
                "infiltration": "Infiltration",
                "interior_equipment": "Equipment",
                "window_energy_flow": "Windows",
                "Wall": "Walls",
                "air_system_heating": "OA Heating",
                "air_system_cooling": "OA Cooling",
                "cooling": "Cooling",
                "heating": "Heating",
            },
            inplace=True,
        )

        heating_load = annual_system_data.xs("Heating Periods", level="Period")
        cooling_load = annual_system_data.xs("Cooling Periods", level="Period")

        end_uses = (
            "Heating",
            "Cooling",
            "Interior Lighting",
            "Exterior Lighting",
            "Interior Equipment",
            "Exterior Equipment",
            "Fans",
            "Pumps",
            "Heat Rejection",
            "Humidification",
            "Heat Recovery",
            "Water Systems",
            "Refrigeration",
            "Generators",
        )
        energy_sources = (
            "Electricity",
            "Natural Gas",
            "District Cooling",
            "District Heating",
        )
        with connect(self.sql_file) as conn:
            df = pd.read_sql(
                'select * from "TabularDataWithStrings" as f where f."TableName" == "End Uses" and f."ReportName" == "AnnualBuildingUtilityPerformanceSummary"',
                conn,
            )
            system_input = df.pivot(
                index="RowName", columns="ColumnName", values="Value"
            ).loc[end_uses, energy_sources]
            system_input = system_input.astype("float")
            system_input = EnergyDataFrame(
                system_input.values,
                index=system_input.index,
                columns=system_input.columns,
            )
            system_input.units = df.set_index("ColumnName").Units.to_dict()
            system_input = system_input.to_units("kWh")

        floor_area = pd.to_numeric(
            Sql(self.sql_file)
            .tabular_data_by_name(
                *(
                    "AnnualBuildingUtilityPerformanceSummary",
                    "Building Area",
                    "Entire Facility",
                )
            )
            .loc["Net Conditioned Building Area", ("Area", "m2")]
        )

        system_input = (
            system_input.replace({0: np.NaN})
            .dropna(how="all")
            .dropna(how="all", axis=1)
        )
        system_input.rename_axis("source", axis=1, inplace=True)
        system_input.rename_axis("target", axis=0, inplace=True)
        system_input = system_input.unstack().rename("value").reset_index().dropna()
        system_input_data = system_input.to_dict(orient="records")

        heating_energy_to_heating_system = [
            {
                "source": "Heating",
                "target": "Heating System",
                "value": system_input.set_index("target").at["Heating", "value"].sum(),
            }
        ]

        cooling_energy_to_heating_system = [
            {
                "source": "Cooling",
                "target": "Cooling System",
                "value": system_input.set_index("target").at["Cooling", "value"].sum(),
            }
        ]

        (
            heating_load_source_data,
            heating_load_target_data,
            link_heating_system_to_gains,
        ) = self._sankey_heating(heating_load, load_type="heating")

        (
            cooling_load_source_data,
            cooling_load_target_data,
            link_cooling_system_to_gains,
        ) = self._sankey_cooling(cooling_load, load_type="cooling")

        flows = pd.DataFrame(
            system_input_data
            + heating_energy_to_heating_system
            + heating_load_source_data
            + heating_load_target_data
            + cooling_energy_to_heating_system
            + cooling_load_source_data
            + cooling_load_target_data
        )
        # Fix the HVAC difference
        diff = (
            flows.loc[flows.source == "Heating Load", "value"].sum()
            - flows.loc[flows.target == "Heating Load", "value"].sum()
        )
        flows.loc[flows.source == "Heating System", "value"] = (
            flows.loc[flows.source == "Heating System", "value"] + diff
        )

        diff = (
            flows.loc[flows.source == "Cooling Load", "value"].sum()
            - flows.loc[flows.target == "Cooling Load", "value"].sum()
        )
        flows.loc[flows.source == "Cooling System", "value"] = (
            flows.loc[flows.source == "Cooling System", "value"] + diff
        )

        # fix names
        flows.replace({"OA Heating Heat Losses": "OA Heating"}, inplace=True)

        # TO EUI
        flows["value"] = flows["value"] / floor_area
        links = pd.DataFrame(
            link_heating_system_to_gains + link_cooling_system_to_gains
        )
        return pd.concat([flows, links]).to_csv(path_or_buf, index=False)

    def _sankey_heating(self, load, load_type="heating"):
        assert load_type in ["heating", "cooling"]
        load_source = (
            load.unstack("Gain/Loss")
            .replace({0: np.NaN})
            .loc[:, "Heat Gain"]
            .dropna(how="all")
            .apply(abs)
            .rename("value")
            .reset_index()
        )
        load_target = (
            load.unstack("Gain/Loss")
            .replace({0: np.NaN})
            .loc[:, "Heat Loss"]
            .dropna(how="all")
            .apply(abs)
            .rename("value")
            .reset_index()
        )
        load_source["target"] = load_type.title() + " Load"
        load_source = load_source.rename({"Component": "source"}, axis=1)
        load_source["source"] = load_source["source"] + " Gain"
        load_source = load_source.replace(
            {f"{load_type.title()} Gain": load_type.title() + " System"}
        )

        load_source_data = load_source.to_dict(orient="records")
        load_target["source"] = load_type.title() + " Load"
        load_target = load_target.rename({"Component": "target"}, axis=1)
        load_target["target"] = load_target["target"] + " Heat Losses"
        load_target_data = load_target.to_dict(orient="records")
        link_system_to_gains = (
            load_source.set_index("source")
            .drop(load_type.title() + " System", errors="ignore")
            .rename_axis("target")
            .apply(lambda x: 0.01, axis=1)
            .rename("value")
            .reset_index()
        )
        link_system_to_gains["source"] = load_type.title()
        link_system_to_gains = link_system_to_gains.to_dict(orient="records")
        return (
            load_source_data,
            load_target_data,
            link_system_to_gains,
        )

    def _sankey_cooling(self, load, load_type="cooling"):
        assert load_type in ["heating", "cooling"]
        load_source = (
            load.unstack("Gain/Loss")
            .replace({0: np.NaN})
            .loc[:, "Heat Loss"]
            .dropna(how="all")
            .apply(abs)
            .rename("value")
            .reset_index()
        )
        load_target = (
            load.unstack("Gain/Loss")
            .replace({0: np.NaN})
            .loc[:, "Heat Gain"]
            .dropna(how="all")
            .apply(abs)
            .rename("value")
            .reset_index()
        )
        load_source["target"] = load_type.title() + " Load"
        load_source = load_source.rename({"Component": "source"}, axis=1)
        load_source["source"] = load_source["source"] + " Losses"
        load_source = load_source.replace(
            {f"{load_type.title()} Losses": load_type.title() + " System"}
        )

        load_source_data = load_source.to_dict(orient="records")
        load_target["source"] = load_type.title() + " Load"
        load_target = load_target.rename({"Component": "target"}, axis=1)
        load_target = (
            load_target.set_index("target")
            .drop(load_type.title(), errors="ignore")
            .reset_index()
        )
        load_target_data = load_target.to_dict(orient="records")
        link_system_to_gains = (
            load_source.set_index("source")
            .drop(load_type.title() + " System", errors="ignore")
            .rename_axis("target")
            .apply(lambda x: 0.01, axis=1)
            .rename("value")
            .reset_index()
        )
        link_system_to_gains["source"] = load_type.title()
        link_system_to_gains = link_system_to_gains.to_dict(orient="records")
        return (
            load_source_data,
            load_target_data,
            link_system_to_gains,
        )



"""Slab module"""

import logging as lg
import shutil
import subprocess
import time
from threading import Thread

from packaging.version import Version
from path import Path
from tqdm.auto import tqdm
from tqdm.contrib.logging import logging_redirect_tqdm

from ..utils import log

from ..eplus_interface.exceptions import EnergyPlusProcessError


class BasementThread(Thread):
    """Basement program manager.

    The basement program used to calculate the results is included with the
    EnergyPlus distribution. It requires an input file named GHTin.idf in
    input data file format. The needed corresponding idd file is
    SlabGHT.idd. An EnergyPlus weather file for the location is also needed.
    """

    def __init__(self, idf, tmp):
        """Constructor."""
        super(BasementThread, self).__init__()
        self.p = None
        self.std_out = None
        self.std_err = None
        self.idf = idf
        self.cancelled = False
        self.run_dir = Path(tmp).expand()
        self.exception = None
        self.name = "RunBasement_" + self.idf.name
        self.include = None

    @property
    def cmd(self):
        """Get the command."""
        cmd_path = Path(shutil.which("Basement", path=self.run_dir))
        return [cmd_path.relpath(self.run_dir)]

    def run(self):
        """Wrapper around the Basement command line interface."""
        self.cancelled = False
        # get version from IDF object or by parsing the IDF file for it

        # Move files into place
        # copy "%wthrfile%.epw" in.epw
        self.epw = self.idf.epw.copy(self.run_dir / "in.epw").expand()
        self.idfname = Path(self.idf.savecopy(self.run_dir / "in.idf")).expand()
        self.idd = self.idf.iddname.copy(self.run_dir).expand()

        # Get executable using shutil.which (determines the extension based on
        # the platform, eg: .exe. And copy the executable to tmp
        basemenet_exe = shutil.which("Basement", path=self.eplus_home)
        if basemenet_exe is None:
            log(
                f"The Basement program could not be found at " f"'{self.eplus_home}",
                lg.WARNING,
            )
            return
        else:
            basemenet_exe = (self.eplus_home / Path(basemenet_exe)).expand()
        self.basement_exe = Path(basemenet_exe).copy(self.run_dir)
        self.basement_idd = (self.eplus_home / "BasementGHT.idd").copy(self.run_dir)
        self.outfile = self.idf.name

        # The BasementGHTin.idf file is copied from the self.include list (
        # added by ExpandObjects. If self.include is empty, no need to run
        # Basement.
        self.include = [Path(file).copy(self.run_dir) for file in self.idf.include]
        if "BasementGHTIn.idf" not in self.include:
            self.cleanup_callback()
            return

        self.msg_callback(
            "===== (Run Basement Temperature Generation) ===== Start ====="
        )
        self.msg_callback("Running Basement.exe")
        self.msg_callback(f"Input File  : {self.idfname}")
        self.msg_callback(
            f"Output Files: {self.outfile}_bsmt.csv "
            f"{self.outfile}_bsmt.audit {self.outfile}_bsmt.out"
        )
        self.msg_callback(f"Weather File: {self.epw}")

        # Run Slab Program
        with logging_redirect_tqdm(loggers=[lg.getLogger(self.idf.name)]):
            with tqdm(
                unit_scale=True,
                miniters=1,
                desc=f"RunBasement #{self.idf.position}-{self.idf.name}",
                position=self.idf.position,
            ) as progress:

                self.p = subprocess.Popen(
                    self.cmd,
                    stdout=subprocess.PIPE,
                    stderr=subprocess.PIPE,
                    shell=True,  # can use shell
                    cwd=self.run_dir,
                )
                start_time = time.time()
                self.msg_callback(
                    "Begin Basement Temperature Calculation processing . . ."
                )

                for line in self.p.stdout:
                    self.msg_callback(line.decode("utf-8").strip("\n"))
                    progress.update()

                # We explicitly close stdout
                self.p.stdout.close()

                # Wait for process to complete
                self.p.wait()

                # Communicate callbacks
                if self.cancelled:
                    self.msg_callback("RunSlab cancelled")
                    # self.cancelled_callback(self.std_out, self.std_err)
                else:
                    if self.p.returncode == 0:
                        self.msg_callback(
                            "RunSlab completed in {:,.2f} seconds".format(
                                time.time() - start_time
                            )
                        )
                        self.success_callback()
                        for line in self.p.stderr:
                            self.msg_callback(line.decode("utf-8"))
                    else:
                        self.msg_callback("RunSlab failed")
                        self.failure_callback()

    def msg_callback(self, *args, **kwargs):
        """Pass message to logger."""
        log(*args, name=self.idf.name, **kwargs)

    def success_callback(self):
        """Parse surface temperature and append to IDF file."""
        csv_ = self.run_dir / "MonthlyResults.csv"
        if csv_.exists():
            csv_ = csv_.rename(self.idf.output_directory / f"{self.outfile}_bsmt.csv")

        input_ = self.run_dir / "RunINPUT.TXT"
        if input_.exists():
            input_ = input_.rename(
                self.idf.output_directory / f"{self.outfile}_bsmt.out"
            )

        debug_ = self.run_dir / "RunDEBUGOUT.txt"
        if debug_.exists():
            debug_ = debug_.rename(self.idf.output_directory / "basementout.audit")

        err_ = self.run_dir / "eplusout.err"
        if err_.exists():
            with open(err_) as f:
                with open(debug_, "a") as f1:
                    for line in f:
                        f1.write(line)

        audit_ = self.run_dir / "audit.out"
        if audit_.exists():
            with open(audit_) as f:
                with open(debug_, "a") as f1:
                    for line in f:
                        f1.write(line)

        ep_objects = self.run_dir / "EPObjects.txt"
        if ep_objects.exists():
            with open(self.idf.idfname, "a") as outfile:
                with open(ep_objects) as infile:
                    next(infile)  # Skipping first line
                    next(infile)  # Skipping second line
                    for line in infile:
                        outfile.write(line)
            # invalidate attributes dependant on idfname, since it has changed
            self.idf._reset_dependant_vars("idfname")
        self.cleanup_callback()

    def cleanup_callback(self):
        """clean up temp files, directories and variables that need cleanup."""

        # Remove from include
        ghtin = self.idf.output_directory / "BasementGHTIn.idf"
        if ghtin.exists():
            try:
                self.idf.include.remove(ghtin)
                ghtin.remove()
            except ValueError:
                log("nothing to remove", lg.DEBUG)

    def failure_callback(self):
        """Parse error file and log"""
        error_filename = self.run_dir / "eplusout.err"
        if error_filename.exists():
            with open(error_filename, "r") as stderr:
                stderr_r = stderr.read()
                self.exception = EnergyPlusProcessError(
                    cmd=self.cmd, stderr=stderr_r, idf=self.idf
                )
        self.cleanup_callback()

    def cancelled_callback(self, stdin, stdout):
        """Call on cancelled."""
        self.cleanup_callback()

    @property
    def eplus_home(self):
        """Get the version-dependant directory where executables are installed."""
        if self.idf.file_version <= Version("7.2"):
            install_dir = self.idf.file_version.current_install_dir / "bin"
        else:
            install_dir = (
                self.idf.file_version.current_install_dir
                / "PreProcess"
                / "GrndTempCalc"
            )
        return install_dir.expand()

    def stop(self):
        if self.p.poll() is None:
            self.msg_callback("Attempting to cancel simulation ...")
            self.cancelled = True
            self.p.kill()



import functools
from operator import add

from validator_collection import validators

from archetypal.template.constructions.opaque_construction import OpaqueConstruction


class InternalMass:
    """InternalMass class."""

    def __init__(self, surface_name, construction, total_area_exposed_to_zone):
        """Create an InternalMass object."""
        self.surface_name = surface_name
        self.construction = construction
        self.total_area_exposed_to_zone = total_area_exposed_to_zone

    @property
    def surface_name(self):
        """Get or set the surface name [string]."""
        return self._surface_name

    @surface_name.setter
    def surface_name(self, value):
        self._surface_name = validators.string(
            value, minimum_length=1, maximum_length=100
        )

    @property
    def construction(self) -> OpaqueConstruction:
        """Get or set the construction."""
        return self._construction

    @construction.setter
    def construction(self, value):
        assert isinstance(value, OpaqueConstruction), (
            f"Input value error for {value}. construction must be of type "
            f"{OpaqueConstruction}, not {type(value)}."
        )
        self._construction = value

    @property
    def total_area_exposed_to_zone(self):
        """Get or set the total area exposed to Zone [m2]."""
        return self._total_area_exposed_to_zone

    @total_area_exposed_to_zone.setter
    def total_area_exposed_to_zone(self, value):
        self._total_area_exposed_to_zone = validators.float(value, minimum=0)

    @classmethod
    def from_zone(cls, zone_epbunch):
        """Create InternalMass from ZoneDefinition and Zone EpBunch.

        Args:
            zone_epbunch (EpBunch): The Zone EpBunch object.

        Returns:
            Construction: The internal mass construction for the zone
            None: if no internal mass defined for zone.
        """
        internal_mass_objs = zone_epbunch.getreferingobjs(
            iddgroups=["Thermal Zones and Surfaces"], fields=["Zone_or_ZoneList_Name"]
        )

        area = 0  # initialize area
        mass_opaque_constructions = []  # collect internal mass objects

        # Looping over possible InternalMass objects
        # This InternalMass object (int_obj) is assigned to self,
        # then create object and append to list. There could be more then
        # one.
        for int_obj in internal_mass_objs:
            if int_obj.key.upper() == "INTERNALMASS":
                mass_opaque_constructions.append(
                    OpaqueConstruction.from_epbunch(int_obj, Category="Internal Mass")
                )
                area += float(int_obj.Surface_Area)

        # If one or more constructions, combine them into one.
        if mass_opaque_constructions:
            # Combine elements and assign the aggregated Surface Area
            construction = functools.reduce(add, mass_opaque_constructions)
        else:
            # No InternalMass object assigned to this Zone, then return Zone and set
            # floor area to 0
            return cls.generic_internalmass_from_zone(zone_epbunch)
        return cls(f"{zone_epbunch.Name} InternalMass", construction, area)

    @classmethod
    def generic_internalmass_from_zone(cls, zone_epbunch):
        """Create an InternalMass object with generic construction and 0 floor area.

        Args:
            zone_epbunch (EpBunch): A ZoneDefinition object.
        """
        construction = OpaqueConstruction.generic_internalmass()
        return cls(
            surface_name=f"{zone_epbunch.Name} InternalMass",
            total_area_exposed_to_zone=0,
            construction=construction,
        )

    def duplicate(self):
        """Get a copy of self."""
        return self.__copy__()

    def mapping(self):
        """Get a dict based on the object properties, useful for dict repr."""
        return {
            "surface_name": self.surface_name,
            "construction": self.construction,
            "total_area_exposed_to_zone": self.total_area_exposed_to_zone,
        }

    def to_epbunch(self, idf, zone_name):
        """Create an `INTERNALMASS` EpBunch given an idf model and a zone name.

        Args:
            idf (IDF): An idf model to add the EpBunch in.
            zone_name (str): The name of the zone for this InternamMass object.

        Returns:
            EpBunch: The EpBunch object added to the idf model.
        """
        construction = self.construction.to_epbunch(idf)
        internal_mass = idf.newidfobject(
            key="INTERNALMASS",
            Name=self.surface_name,
            Construction_Name=construction.Name,
            Zone_or_ZoneList_Name=zone_name,
            Surface_Area=self.total_area_exposed_to_zone,
        )
        return internal_mass

    def __copy__(self):
        """Get a copy of self."""
        return self.__class__(**self.mapping())

    def __eq__(self, other):
        """Assert self equals to other."""
        return isinstance(other, InternalMass) and self.__key__() == other.__key__()

    def __key__(self):
        """Get a tuple of attributes. Useful for hashing and comparing."""
        return tuple(self.mapping().values())



"""EnergyPlus reports module."""

import os
from sqlite3.dbapi2 import OperationalError

import pandas as pd

from archetypal import settings
from archetypal.idfclass.util import hash_model
from archetypal.utils import log


def get_report(
    eplus_file, output_directory=None, output_report="sql", output_prefix=None, **kwargs
):
    """Returns the specified report format (html or sql)

    Args:
        eplus_file (str): path of the idf file
        output_directory (str, optional): path to the output folder. Will
            default to the settings.cache_folder.
        output_report: 'htm' or 'sql'
        output_prefix (str): Prefix name given to results files.
        **kwargs: keyword arguments to pass to hasher.

    Returns:
        dict: a dict of DataFrames
    """
    if not output_directory:
        output_directory = settings.cache_folder
    # Hash the idf file with any kwargs used in the function
    if output_prefix is None:
        output_prefix = hash_model(eplus_file)
    if output_report is None:
        return None
    elif "htm" in output_report.lower():
        # Get the html report
        fullpath_filename = output_directory / output_prefix + "tbl.htm"
        if fullpath_filename.exists():
            return get_html_report(fullpath_filename)
        else:
            raise FileNotFoundError(
                'File "{}" does not exist'.format(fullpath_filename)
            )

    elif "sql" == output_report.lower():
        # Get the sql report
        fullpath_filename = output_directory / output_prefix + "out.sql"
        if fullpath_filename.exists():
            return get_sqlite_report(fullpath_filename)
        else:
            raise FileNotFoundError(
                'File "{}" does not exist'.format(fullpath_filename)
            )
    else:
        return None


def get_html_report(report_fullpath):
    """Parses the html Summary Report for each tables into a dictionary of
    DataFrames

    Args:
        report_fullpath (str): full path to the report file

    Returns:
        dict: dict of {title : table <DataFrame>,...}
    """
    from eppy.results import readhtml  # the eppy module with functions to read the html

    with open(report_fullpath, "r", encoding="utf-8") as cache_file:
        filehandle = cache_file.read()  # get a file handle to the html file

        cached_tbl = readhtml.titletable(
            filehandle
        )  # get a file handle to the html file

        log('Retrieved response from cache file "{}"'.format(report_fullpath))
        return summary_reports_to_dataframes(cached_tbl)


def summary_reports_to_dataframes(reports_list):
    """Converts a list of [(title, table),...] to a dict of {title: table
    <DataFrame>}. Duplicate keys must have their own unique names in the output
    dict.

    Args:
        reports_list (list): a list of [(title, table),...]

    Returns:
        dict: a dict of {title: table <DataFrame>}
    """
    results_dict = {}
    for table in reports_list:
        key = str(table[0])
        if key in results_dict:  # Check if key is already exists in
            # dictionary and give it a new name
            key = key + "_"
        df = pd.DataFrame(table[1])
        df = df.rename(columns=df.iloc[0]).drop(df.index[0])
        results_dict[key] = df
    return results_dict


def get_sqlite_report(report_file, report_tables=None):
    """Connect to the EnergyPlus SQL output file and retrieves all tables

    Args:
        report_file (str): path of report file
        report_tables (list, optional): list of report table names to retreive.
            Defaults to settings.available_sqlite_tables

    Returns:
        dict: dict of DataFrames
    """
    # set list of report tables
    if not report_tables:
        report_tables = settings.available_sqlite_tables

    # if file exists, parse it with pandas' read_sql_query
    if os.path.isfile(report_file):
        import sqlite3

        import numpy as np

        # create database connection with sqlite3
        with sqlite3.connect(report_file) as conn:
            # empty dict to hold all DataFrames
            all_tables = {}
            # Iterate over all tables in the report_tables list
            for table in report_tables:
                try:
                    # Try regular str read, could fail if wrong encoding
                    conn.text_factory = str
                    df = pd.read_sql_query(
                        "select * from {};".format(table),
                        conn,
                        index_col=report_tables[table]["PrimaryKey"],
                        parse_dates=report_tables[table]["ParseDates"],
                        coerce_float=True,
                    )
                    all_tables[table] = df
                except OperationalError:
                    # Wring encoding found, the load bytes and ecode object
                    # columns only
                    conn.text_factory = bytes
                    df = pd.read_sql_query(
                        "select * from {};".format(table),
                        conn,
                        index_col=report_tables[table]["PrimaryKey"],
                        parse_dates=report_tables[table]["ParseDates"],
                        coerce_float=True,
                    )
                    str_df = df.select_dtypes([np.object])
                    str_df = str_df.stack().str.decode("8859").unstack()
                    for col in str_df:
                        df[col] = str_df[col]
                    all_tables[table] = df
            log(
                "SQL query parsed {} tables as DataFrames from {}".format(
                    len(all_tables), report_file
                )
            )
            return all_tables


def get_ideal_loads_summary(idf):
    """Get ideal loads report summary.

    Returns DataFrame with Cooling, Domestic Hot Water, Equipment, Heating and
    Lighting hourly time series.

    Returns:
        DataFrame: DataFrame of hourly time series.
    """
    res = {
        "SDL/Cooling": idf.meters.OutputMeter.Cooling__DistrictCooling.values("kWh"),
        "SDL/Domestic Hot Water": idf.meters.OutputMeter.WaterSystems__DistrictHeating.values(
            "kWh"
        ),
        "SDL/Equipment": idf.meters.OutputMeter.InteriorEquipment__Electricity.values(
            "kWh"
        ),
        "SDL/Heating": idf.meters.OutputMeter.Heating__DistrictHeating.values("kWh"),
        "SDL/Lighting": idf.meters.OutputMeter.InteriorLights__Electricity.values(
            "kWh"
        ),
    }
    return pd.DataFrame(res)



"""UmiTemplateLibrary Module."""

import json
import logging as lg
from collections import OrderedDict
from concurrent.futures.thread import ThreadPoolExecutor
from typing import List

import networkx as nx
from pandas.io.common import get_handle
from path import Path

from archetypal import settings
from archetypal.eplus_interface.exceptions import EnergyPlusProcessError
from archetypal.idfclass.idf import IDF
from archetypal.template.building_template import BuildingTemplate
from archetypal.template.conditioning import ZoneConditioning
from archetypal.template.constructions.opaque_construction import OpaqueConstruction
from archetypal.template.constructions.window_construction import WindowConstruction
from archetypal.template.dhw import DomesticHotWaterSetting
from archetypal.template.load import ZoneLoad
from archetypal.template.materials.gas_layer import GasLayer
from archetypal.template.materials.gas_material import GasMaterial
from archetypal.template.materials.glazing_material import GlazingMaterial
from archetypal.template.materials.material_layer import MaterialLayer
from archetypal.template.materials.opaque_material import OpaqueMaterial
from archetypal.template.schedule import (
    DaySchedule,
    UmiSchedule,
    WeekSchedule,
    YearSchedule,
    YearSchedulePart,
)
from archetypal.template.structure import MassRatio, StructureInformation
from archetypal.template.umi_base import UmiBase, UniqueName
from archetypal.template.ventilation import VentilationSetting
from archetypal.template.window_setting import WindowSetting
from archetypal.template.zone_construction_set import ZoneConstructionSet
from archetypal.template.zonedefinition import ZoneDefinition
from archetypal.utils import CustomJSONEncoder, log, parallel_process


class UmiTemplateLibrary:
    """Handles parsing and creating Template Library Files for UMI for Rhino.

    - See :meth:`open` to parse existing Umi Template Library files (.json).
    - See :meth:`from_idf_files` to create a library by converting existing IDF models.
    """

    _LIB_GROUPS = [
        "GasMaterials",
        "GlazingMaterials",
        "OpaqueMaterials",
        "OpaqueConstructions",
        "WindowConstructions",
        "StructureInformations",
        "DaySchedules",
        "WeekSchedules",
        "YearSchedules",
        "DomesticHotWaterSettings",
        "VentilationSettings",
        "ZoneConditionings",
        "ZoneConstructionSets",
        "ZoneLoads",
        "ZoneDefinitions",
        "WindowSettings",
        "BuildingTemplates",
    ]

    def __init__(
        self,
        name="unnamed",
        BuildingTemplates=None,
        GasMaterials=None,
        GlazingMaterials=None,
        OpaqueConstructions=None,
        OpaqueMaterials=None,
        WindowConstructions=None,
        StructureInformations=None,
        DaySchedules=None,
        WeekSchedules=None,
        YearSchedules=None,
        DomesticHotWaterSettings=None,
        VentilationSettings=None,
        WindowSettings=None,
        ZoneConditionings=None,
        ZoneConstructionSets=None,
        ZoneLoads=None,
        ZoneDefinitions=None,
    ):
        """Initialize a new UmiTemplateLibrary with empty attributes.

        Args:
            name (str): The name of the UMI Template.
            BuildingTemplates (list of BuildingTemplate): list of
                BuildingTemplate objects.
            GasMaterials (list of GasMaterial): list of GasMaterial objects.
            GlazingMaterials (list of GlazingMaterial): list of GlazingMaterial
                objects.
            OpaqueConstructions (list of OpaqueConstruction): list of
                OpaqueConstruction objects.
            OpaqueMaterials (list of OpaqueMaterial): list of OpaqueMaterial
                objects.
            WindowConstructions (list of WindowConstruction): list of
                WindowConstruction objects.
            StructureInformations (list of StructureInformation): list of
                StructureInformation objects.
            DaySchedules (list of DaySchedule): list of DaySchedule objects.
            WeekSchedules (list of WeekSchedule): list of WeekSchedule objects.
            YearSchedules (list of YearSchedule): list of YearSchedule objects.
            DomesticHotWaterSettings (list of DomesticHotWaterSetting): list of
                DomesticHotWaterSetting objects.
            VentilationSettings (list of VentilationSetting): list of
                VentilationSetting objects.
            WindowSettings (list of WindowSetting): list of WindowSetting
                objects.
            ZoneConditionings (list of ZoneConditioning): list of
                ZoneConditioning objects.
            ZoneConstructionSets (list of ZoneConstructionSet): list of
                ZoneConstructionSet objects.
            ZoneLoads (list of ZoneLoad): list of ZoneLoad objects.
            ZoneDefinitions (list of ZoneDefinition): list of Zone objects
        """
        self.idf_files = []
        self.name = name
        self.ZoneDefinitions = ZoneDefinitions or []
        self.ZoneLoads = ZoneLoads or []
        self.ZoneConstructionSets = ZoneConstructionSets or []
        self.ZoneConditionings = ZoneConditionings or []
        self.WindowSettings = WindowSettings or []
        self.VentilationSettings = VentilationSettings or []
        self.DomesticHotWaterSettings = DomesticHotWaterSettings or []
        self.UmiSchedules = []  # placeholder for UmiSchedules
        self.YearSchedules = YearSchedules or []
        self.WeekSchedules = WeekSchedules or []
        self.DaySchedules = DaySchedules or []
        self.StructureInformations = StructureInformations or []
        self.WindowConstructions = WindowConstructions or []
        self.OpaqueMaterials = OpaqueMaterials or []
        self.OpaqueConstructions = OpaqueConstructions or []
        self.BuildingTemplates = BuildingTemplates or []
        self.GasMaterials = GasMaterials or []
        self.GlazingMaterials = GlazingMaterials or []

    def __iter__(self):
        """Iterate over component groups. Yields tuple of (group, value)."""
        for group in self._LIB_GROUPS:
            yield group, self.__dict__[group]

    def __getitem__(self, item):
        return self.__dict__[item]

    def __add__(self, other: "UmiTemplateLibrary"):
        """Combined"""
        for key, group in other:
            # for each group items
            for component in group:
                component.id = None  # Reset the component's id

        attrs = {}
        for group, value in self:
            attrs[group] = value + other.__dict__[group]

        newlib = self.__class__(**attrs, name=self.name)
        newlib.unique_components("GasMaterials", keep_orphaned=True)
        return newlib

    def _clear_components_list(self, except_groups=None):
        """Clear components lists except except_groups."""
        if except_groups is None:
            except_groups = []
        exception = ["BuildingTemplates"]
        exception.extend(except_groups)
        for key, group in self:
            if key not in exception:
                setattr(self, key, [])

    @property
    def object_list(self):
        """Get list of all objects in self, including orphaned objects."""
        objs = []
        for name, group in self:
            objs.extend(group)
        return objs

    @classmethod
    def from_idf_files(
        cls,
        idf_files,
        weather,
        name="unnamed",
        processors=-1,
        keep_all_zones=False,
        unique_components=None,
        debug=False,
        **kwargs,
    ):
        """Initialize an UmiTemplateLibrary object from one or more idf_files.

        The resulting object contains the reduced version of the IDF files.
        To save to file, call the :meth:`save` method.

        Important:
            When using :meth:`from_idf_files` The idf files are striped of run period
            modifiers and special days to return simple annual schedules.

        Args:
            idf_files (list of (str or Path)): list of IDF file paths.
            weather (str or Path): Path to the weather file.
            name (str): The name of the Template File
            processors (int): Number of cores. Defaults to -1, all cores.
            debug (bool): If True, will raise any error on any processed file.
            kwargs: keyword arguments passed to IDF().

        Raises:
            Exception: All exceptions are raised if settings.debug=True. Will raise
                an exception if all BuildingTemplates failed to be created.
        """
        # instantiate class
        umi_template = cls(name)

        # if parallel is True, run eplus in parallel
        in_dict = {}
        for i, idf_file in enumerate(idf_files):
            in_dict[idf_file] = dict(
                idfname=idf_file,
                epw=weather,
                verbose=False,
                position=i,
                nolimit=True,
                keep_data_err=True,  # For debugging
                readvars=False,  # No need to readvars since only sql is used
                **kwargs,
            )
        results = parallel_process(
            in_dict,
            cls.template_complexity_reduction,
            processors=processors,
            use_kwargs=True,
            debug=debug,
            position=None,
            executor=ThreadPoolExecutor,
        )
        for filename, res in results.items():
            if isinstance(res, EnergyPlusProcessError):
                filename = (settings.logs_folder / "failed_reduce.txt").expand()
                with open(filename, "a") as file:
                    file.writelines(res.write())
                    log(
                        f"EnergyPlusProcess error for {filename} listed in"
                        f" {filename}: {res}",
                        lg.ERROR,
                    )
            elif isinstance(res, Exception):
                if debug:
                    raise res
                else:
                    log(
                        f"Exception raised for {filename}: {res}",
                        lg.ERROR,
                    )

        # If all exceptions, raise them for debugging
        if all(isinstance(x, Exception) for x in results.values()):
            raise Exception(
                [res for res in results.values() if isinstance(res, Exception)]
            )

        umi_template.BuildingTemplates = [
            res for res in results.values() if not isinstance(res, Exception)
        ]

        if keep_all_zones:
            _zones = set(obj.get_unique() for obj in ZoneDefinition._CREATED_OBJECTS)
            for zone in _zones:
                umi_template.ZoneDefinitions.append(zone)
            exceptions = [ZoneDefinition.__name__]
        else:
            exceptions = None

        # Get unique instances
        umi_template.unique_components(
            *(unique_components or []), exceptions=exceptions
        )

        # Update attributes of instance
        umi_template.update_components_list(exceptions=exceptions)

        return umi_template

    @staticmethod
    def template_complexity_reduction(idfname, epw, **kwargs):
        """Wrap IDF, simulate and BuildingTemplate for parallel processing."""
        idf = IDF(idfname, epw=epw, **kwargs)
        idf._outputs.add_umi_template_outputs()

        # remove daylight saving time modifiers
        for daylight in idf.idfobjects["RunPeriodControl:DaylightSavingTime".upper()]:
            idf.removeidfobject(daylight)
        # edit run period to start on Monday
        for run_period in idf.idfobjects["RunPeriod".upper()]:
            run_period.Day_of_Week_for_Start_Day = "Monday"
            run_period.Apply_Weekend_Holiday_Rule = "No"
            run_period.Use_Weather_File_Holidays_and_Special_Days = "No"
            run_period.Use_Weather_File_Daylight_Saving_Period = "No"
        # remove daylight saving time modifiers
        for day in idf.idfobjects["RunPeriodControl:SpecialDays".upper()]:
            idf.removeidfobject(day)

        try:
            idf.simulate()
        except EnergyPlusProcessError as e:
            return e
        return BuildingTemplate.from_idf(idf, **kwargs)

    @classmethod
    def open(cls, filename):
        """Initialize an UmiTemplate object from an UMI Template Library File.

        Args:
            filename (str or Path): PathLike object giving the pathname of the UMI
                Template File.

        Returns:
            UmiTemplateLibrary: The template object.
        """
        name = Path(filename)
        with open(filename, "r") as f:
            t = cls.loads(f.read(), name)

        return t

    @classmethod
    def loads(cls, s, name):
        """load string."""
        datastore = json.loads(s)
        # with datastore, create each objects
        t = cls(name)
        t.GasMaterials = [
            GasMaterial.from_dict(store, allow_duplicates=False)
            for store in datastore["GasMaterials"]
        ]
        t.GlazingMaterials = [
            GlazingMaterial.from_dict(store, allow_duplicates=False)
            for store in datastore["GlazingMaterials"]
        ]
        t.OpaqueMaterials = [
            OpaqueMaterial.from_dict(store, allow_duplicates=False)
            for store in datastore["OpaqueMaterials"]
        ]
        t.OpaqueConstructions = [
            OpaqueConstruction.from_dict(
                store,
                materials={
                    a.id: a
                    for a in (t.GasMaterials + t.GlazingMaterials + t.OpaqueMaterials)
                },
                allow_duplicates=True,
            )
            for store in datastore["OpaqueConstructions"]
        ]
        t.WindowConstructions = [
            WindowConstruction.from_dict(
                store,
                materials={a.id: a for a in (t.GasMaterials + t.GlazingMaterials)},
                allow_duplicates=True,
            )
            for store in datastore["WindowConstructions"]
        ]
        t.StructureInformations = [
            StructureInformation.from_dict(
                store,
                materials={a.id: a for a in t.OpaqueMaterials},
                allow_duplicates=True,
            )
            for store in datastore["StructureDefinitions"]
        ]
        t.DaySchedules = [
            DaySchedule.from_dict(store, allow_duplicates=True)
            for store in datastore["DaySchedules"]
        ]
        t.WeekSchedules = [
            WeekSchedule.from_dict(
                store,
                day_schedules={a.id: a for a in t.DaySchedules},
                allow_duplicates=True,
            )
            for store in datastore["WeekSchedules"]
        ]
        t.YearSchedules = [
            YearSchedule.from_dict(
                store,
                week_schedules={a.id: a for a in t.WeekSchedules},
                allow_duplicates=True,
            )
            for store in datastore["YearSchedules"]
        ]
        t.DomesticHotWaterSettings = [
            DomesticHotWaterSetting.from_dict(
                store,
                schedules={a.id: a for a in t.YearSchedules},
                allow_duplicates=True,
            )
            for store in datastore["DomesticHotWaterSettings"]
        ]
        t.VentilationSettings = [
            VentilationSetting.from_dict(
                store,
                schedules={a.id: a for a in t.YearSchedules},
                allow_duplicates=True,
            )
            for store in datastore["VentilationSettings"]
        ]
        t.ZoneConditionings = [
            ZoneConditioning.from_dict(
                store,
                schedules={a.id: a for a in t.YearSchedules},
                allow_duplicates=True,
            )
            for store in datastore["ZoneConditionings"]
        ]
        t.ZoneConstructionSets = [
            ZoneConstructionSet.from_dict(
                store,
                opaque_constructions={a.id: a for a in t.OpaqueConstructions},
                allow_duplicates=True,
            )
            for store in datastore["ZoneConstructionSets"]
        ]
        t.ZoneLoads = [
            ZoneLoad.from_dict(
                store,
                schedules={a.id: a for a in t.YearSchedules},
                allow_duplicates=True,
            )
            for store in datastore["ZoneLoads"]
        ]
        t.ZoneDefinitions = [
            ZoneDefinition.from_dict(
                store,
                zone_conditionings={a.id: a for a in t.ZoneConditionings},
                zone_construction_sets={a.id: a for a in t.ZoneConstructionSets},
                domestic_hot_water_settings={
                    a.id: a for a in t.DomesticHotWaterSettings
                },
                opaque_constructions={a.id: a for a in t.OpaqueConstructions},
                zone_loads={a.id: a for a in t.ZoneLoads},
                ventilation_settings={a.id: a for a in t.VentilationSettings},
                allow_duplicates=True,
            )
            for store in datastore["Zones"]
        ]
        t.WindowSettings = [
            WindowSetting.from_ref(
                store["$ref"],
                datastore["BuildingTemplates"],
                schedules={a.id: a for a in t.YearSchedules},
                window_constructions={a.id: a for a in t.WindowConstructions},
            )
            if "$ref" in store
            else WindowSetting.from_dict(
                store,
                schedules={a.id: a for a in t.YearSchedules},
                window_constructions={a.id: a for a in t.WindowConstructions},
                allow_duplicates=True,
            )
            for store in datastore["WindowSettings"]
        ]
        t.BuildingTemplates = [
            BuildingTemplate.from_dict(
                store,
                zone_definitions={a.id: a for a in t.ZoneDefinitions},
                structure_definitions={a.id: a for a in t.StructureInformations},
                window_settings={a.id: a for a in t.WindowSettings},
                schedules={a.id: a for a in t.YearSchedules},
                window_constructions={a.id: a for a in t.WindowConstructions},
                allow_duplicates=True,
            )
            for store in datastore["BuildingTemplates"]
        ]
        return t

    def validate(self, defaults=True):
        """Validate the object."""
        pass

    def save(
        self,
        path_or_buf=None,
        indent=2,
        sort_keys=False,
        compression="infer",
        storage_options=None,
    ):
        """Save to json file.

        Args:
            path_or_buf (path-like): File path or object. If not specified,
                overwrites files. See :attr:`UmiTemplateLibrary.name`.
            indent (bool or str or int): If indent is a non-negative integer or string,
                then JSON array elements and object members will be
                pretty-printed with that indent level. An indent level of 0,
                negative, or "" will only insert newlines. None (the default)
                selects the most compact representation. Using a positive
                integer indent indents that many spaces per level. If indent is
                a string (such as "t"), that string is used to indent each
                level.
            sort_keys (callable): If sort_keys is true (default: False), then the
                output of dictionaries will be sorted by this callable.
                e.g.: `lambda x: x.get("$id")` sorts by $id. If callable is not
                available or fails, then sorted by `Name`.
            compression (str): A string representing the compression to use in the
                output file, only used when the first argument is a filename. By
                default, the compression is inferred from the filename.
            storage_options (dict): Extra options that make sense for a particular
                storage connection, e.g. host, port, username, password, etc.,
                if using a URL that will be parsed by fsspec, e.g., starting “s3://”,
                “gcs://”. An error will be raised if providing this argument with a
                non-fsspec URL. See the fsspec and backend storage implementation
                docs for the set of allowed keys and values.
        """
        if path_or_buf is None:
            basedir = Path(self.name).dirname()
            name = Path(self.name).stem
            path_or_buf = basedir / name + ".json"

        self.to_json(
            path_or_buf,
            indent=indent,
            sort_keys=sort_keys,
            compression=compression,
            storage_options=storage_options,
        )

    def to_json(
        self,
        path_or_buf=None,
        indent=2,
        sort_keys=False,
        default_handler=None,
        compression="infer",
        storage_options=None,
    ):
        """Convert the object to a JSON string.

        Args:
            path_or_buf (path-like): File path or object. If not specified,
                the result is returned as a string.
            indent (bool or str or int): If indent is a non-negative integer or string,
                then JSON array elements and object members will be
                pretty-printed with that indent level. An indent level of 0,
                negative, or "" will only insert newlines. None (the default)
                selects the most compact representation. Using a positive
                integer indent indents that many spaces per level. If indent is
                a string (such as "t"), that string is used to indent each
                level.
            sort_keys (callable):
            default_handler (callable): Handler to call if object cannot otherwise be
                converted to a suitable format for JSON. Should receive a single
                argument which is the object to convert and return a serializable
                object.
            compression (str): A string representing the compression to use in the
                output file, only used when the first argument is a filename. By
                default, the compression is inferred from the filename.
            storage_options (dict): Extra options that make sense for a particular
                storage connection, e.g. host, port, username, password, etc.,
                if using a URL that will be parsed by fsspec, e.g., starting “s3://”,
                “gcs://”. An error will be raised if providing this argument with a
                non-fsspec URL. See the fsspec and backend storage implementation
                docs for the set of allowed keys and values.
        """
        if default_handler is None:
            default_handler = CustomJSONEncoder

        data_dict = self.to_dict()

        if sort_keys is not None:
            # Sort values
            for key in data_dict:
                # Sort the list elements by their Name
                try:
                    data_dict[key] = sorted(data_dict[key], key=sort_keys)
                except Exception:
                    # revert to sorting by Name if failure
                    data_dict[key] = sorted(data_dict[key], key=lambda x: x.get("Name"))

        response = json.dumps(data_dict, indent=indent, cls=default_handler)
        if path_or_buf is not None:
            # apply compression and byte/text conversion
            with get_handle(
                path_or_buf,
                "wt",
                compression=compression,
                storage_options=storage_options,
            ) as handles:
                handles.handle.write(response)
        else:
            return response

    def to_dict(self):
        """Return UmiTemplateLibrary dictionary representation."""
        # First, reset existing name

        # Create ordered dict with empty list
        data_dict = OrderedDict([(key, []) for key in self._LIB_GROUPS])

        # create dict values
        for group_name, group in self:
            # reset unique names for group
            UniqueName.existing = {}
            obj: UmiBase
            for obj in group:
                try:
                    data = obj.to_dict()
                except AttributeError as e:
                    raise AttributeError(f"{e} for {obj}")
                data.update({"Name": UniqueName(data.get("Name"))})
                data_dict.setdefault(group_name, []).append(data)

        if not data_dict.get("GasMaterials"):
            # Umi needs at least one gas material even if it is not necessary.
            data = GasMaterial(Name="AIR").to_dict()
            data.update({"Name": UniqueName(data.get("Name"))})
            data_dict.get("GasMaterials").append(data)
            data_dict.move_to_end("GasMaterials", last=False)

        # Correct naming convention and reorder categories
        for key in tuple(data_dict.keys()):
            v = data_dict[key]
            del data_dict[key]
            if key == "ZoneDefinitions":
                key = "Zones"
            if key == "StructureInformations":
                key = "StructureDefinitions"
            data_dict[key] = v

        # Validate
        assert no_duplicates(data_dict, attribute="Name")

        # Sort values
        for key in data_dict:
            # Sort the list elements by their Name
            data_dict[key] = sorted(data_dict[key], key=lambda x: x.get("Name"))

        return data_dict

    def unique_components(
        self, *args: str, exceptions: List[str] = None, keep_orphaned=False
    ):
        """Keep only unique components.

        Starts by clearing all objects in self except self.BuildingTemplates.
        Then, recursively traverses the children of each BuildingTemplate, finding a
        unique object for "equivalent" components.

        Calls :func:`~archetypal.template.umi_base.UmiBase.get_unique` for each
        object in the graph.

        Args:
            *args (str): UmiBase class names that should be replaced with a
                unique equivalent. For example, if only "OpaqueMaterials" should be
                unique, then use self.unique_components("OpaqueMaterials"). If none
                are provided, all umi components be unique.
            exceptions (List[str]): A list of UmiBase class names that will not be
                cleared from self.
            keep_orphaned (bool): if True, orphaned objects are kept.
        """
        if keep_orphaned:
            G = self.to_graph(include_orphans=True)
            connected_to_building = set()
            for bldg in self.BuildingTemplates:
                for obj in nx.dfs_preorder_nodes(G, bldg):
                    connected_to_building.add(obj)
            orphans = [
                obj for obj in self.object_list if obj not in connected_to_building
            ]
        self._clear_components_list(exceptions)  # First clear components

        # Inclusion is a set of object classes that will be unique.
        inclusion = set(args or [])
        if not inclusion.isdisjoint(set(self._LIB_GROUPS)):
            inclusion = set(self._LIB_GROUPS).intersection(inclusion)
        else:
            if inclusion:
                assert inclusion.intersection(set(self._LIB_GROUPS)), (
                    f"{', '.join(inclusion.difference(set(self._LIB_GROUPS)))} not a "
                    f"valid class name. Valid values are: "
                    f"{', '.join(set(self._LIB_GROUPS))}"
                )
            inclusion = set(self._LIB_GROUPS)
        for key, group in self:
            # for each group
            for component in group:
                # travers each object using generator
                for parent, key, obj in parent_key_child_traversal(component):
                    if obj.__class__.__name__ + "s" in inclusion:
                        if key:
                            setattr(
                                parent, key, obj.get_unique()
                            )  # set unique object on key

        self.update_components_list(exceptions=exceptions)  # Update the components list
        if keep_orphaned:
            for obj in orphans:
                self[obj.__class__.__name__ + "s"].append(obj)

    def replace_component(self, this, that) -> None:
        """Replace all instances of `this` with `that`.

        Args:
            this (UmiBase): The reference to replace with `that`.
            that (UmiBase): The object to replace each references with.
        """
        for bldg in self.BuildingTemplates:
            for parent, key, obj in parent_key_child_traversal(bldg):
                if obj is this:
                    setattr(parent, key, that)

        self.update_components_list()

    def update_components_list(self, exceptions=None):
        """Update the component groups with connected components."""
        # clear components list except BuildingTemplate
        self._clear_components_list(exceptions)

        for key, group in self:
            for component in group:
                for parent, key, child in parent_key_child_traversal(component):
                    if isinstance(child, UmiSchedule) and not isinstance(
                        child, (DaySchedule, WeekSchedule, YearSchedule)
                    ):
                        y, ws, ds = child.to_year_week_day()
                        if not any(o.id == y.id for o in self.YearSchedules):
                            self.YearSchedules.append(y)
                        for w in ws:
                            if not any(o.id == w.id for o in self.WeekSchedules):
                                self.WeekSchedules.append(w)
                        for d in ds:
                            if not any(o.id == d.id for o in self.DaySchedules):
                                self.DaySchedules.append(d)
                        # finally, replace it with y
                        setattr(parent, key, y)
                    elif isinstance(child, UmiBase):
                        obj_list = self.__dict__[child.__class__.__name__ + "s"]
                        if not any(o.id == child.id for o in obj_list):
                            # Important to compare on UmiBase.id and not on identity.
                            obj_list.append(child)

    def to_graph(self, include_orphans=False):
        """Create a :class:`networkx.DiGraph` of self.

        This networkx.DiGraph object is then useful for graph-theory operations on
        the hierarchy of the UmiTemplateLibrary.
        """
        import networkx as nx

        G = nx.DiGraph()

        for bldg in self.BuildingTemplates:
            for parent, child in parent_child_traversal(bldg):
                if parent:
                    G.add_edge(parent, child)

        if include_orphans:
            orphans = [
                obj for obj in self.object_list if obj.id not in (n.id for n in G)
            ]
            for orphan in orphans:
                G.add_node(orphan)
                for parent, child in parent_child_traversal(orphan):
                    if parent:
                        G.add_edge(parent, child)

        return G


def no_duplicates(file, attribute="Name"):
    """Assert whether or not dict has duplicated Names.

    `attribute` can be another attribute name like "$id".

    Args:
        file (str or dict): Path of the json file or dict containing umi objects groups
        attribute (str): Attribute to search for duplicates in json UMI structure.
            eg. : "$id", "Name".

    Returns:
        bool: True if no duplicates.

    Raises:
        Exception if duplicates found.
    """
    import json
    from collections import defaultdict

    if isinstance(file, str):
        data = json.loads(open(file).read())
    else:
        data = file
    ids = {}
    for key, value in data.items():
        ids[key] = defaultdict(int)
        for component in value:
            try:
                _id = component[attribute]
            except KeyError:
                pass  # BuildingTemplate does not have an id
            else:
                ids[key][_id] += 1
    dups = {
        key: dict(filter(lambda x: x[1] > 1, values.items()))
        for key, values in ids.items()
        if dict(filter(lambda x: x[1] > 1, values.items()))
    }
    if any(dups.values()):
        raise Exception(f"Duplicate {attribute} found: {dups}")
    else:
        return True


DEEP_OBJECTS = (UmiBase, MaterialLayer, GasLayer, YearSchedulePart, MassRatio, list)


def parent_key_child_traversal(parent):
    """Iterate over children of parent yield the parent, the attribute name of the
    children (key) and the children itself.

    If called on a :class:BuildingTemplate object, this will traverse all the
    umi components this object references, like a depth-first search algotrithm in a
    graph structure.
    """
    if isinstance(parent, DEEP_OBJECTS):
        if isinstance(parent, list):
            for obj in parent:
                yield from parent_key_child_traversal(obj)
        elif isinstance(parent, DaySchedule):
            yield None, None, parent
        else:
            for k, child in parent:
                if isinstance(child, UmiBase):
                    yield parent, k, child
                if isinstance(child, DEEP_OBJECTS):
                    yield from parent_key_child_traversal(child)


def parent_child_traversal(parent: UmiBase):
    """Iterate over all children of the parent.

    This generator recursively yields (parent, child) tuples. It uses the
    :attr:`UmiBase.children` attribute and had a better performance than
    :func:`parent_key_child_traversal`.
    """
    for child in parent.children:
        yield parent, child
        yield from parent_child_traversal(child)


def traverse(parent):
    return parent_child_traversal(parent)



import pytest

from archetypal.eplus_interface.exceptions import InvalidEnergyPlusVersion
from archetypal.eplus_interface.version import EnergyPlusVersion


class TestEnergyPlusVersion:
    """"""

    def test_eplusversion_version_init(self):
        """Test the initialization of VersionIdd and basic properties"""
        eplus_version = EnergyPlusVersion("9.2")
        str(eplus_version)  # test the string representation
        assert repr(eplus_version) == "<EnergyPlusVersion('9.2')>"  # test the repr

        assert eplus_version.dash == "9-2-0"
        assert eplus_version.tuple == (9, 2, 0)
        assert isinstance(eplus_version.valid_versions, set)

        # initialized 3 different ways, should return equivalent objects.
        assert (
            EnergyPlusVersion("9.2.0")
            == EnergyPlusVersion((9, 2, 0))
            == EnergyPlusVersion("9-2-0")
        )

        copy_of_idd_version = eplus_version.duplicate()
        assert copy_of_idd_version == eplus_version

        # initialize current module version
        eplus_version = EnergyPlusVersion.current()
        str(eplus_version)  # test the string representation

        # initialize current version
        eplus_version = EnergyPlusVersion.latest()
        str(eplus_version)  # test the string representation

        assert isinstance(eplus_version.current_idd_path, str)

    def test_eplusversion_err_init(self):
        """Test initialization of VersionIDd which produces an error."""
        with pytest.raises(InvalidEnergyPlusVersion):
            idd_version = EnergyPlusVersion("3.2.0")

    def test_idd_on_missing_install(self):
        idd_version = EnergyPlusVersion("9.2")
        idd_version._valid_paths = {}  # fakes not finding any versions on machine.

        assert idd_version.valid_versions == {"9-2-0",}
